=== added file 'DBUS-API'
--- DBUS-API	1970-01-01 00:00:00 +0000
+++ DBUS-API	2010-09-12 03:00:40 +0000
@@ -0,0 +1,147 @@
+		   -*- mode: org; coding: utf-8 -*-
+
+		    Mandos Server D-Bus Interface
+
+This file documents the D-Bus interface to the Mandos server.
+
+* Bus: System bus
+  Bus name: "se.bsnet.fukt.Mandos"
+
+
+* Object Paths:
+  
+  | Path                  | Object            |
+  |-----------------------+-------------------|
+  | "/"                   | The Mandos Server |
+  | "/clients/CLIENTNAME" | Mandos Client     |
+
+  
+* Mandos Server Interface:
+  Interface name: "se.bsnet.fukt.Mandos"
+  
+** Methods:
+*** GetAllClients() → (ao: Clients)
+    Returns an array of all client D-Bus object paths
+   
+*** GetAllClientsWithProperties() → (a{oa{sv}}: ClientProperties)
+    Returns an array of all clients and all their properties
+   
+*** RemoveClient(o: ObjectPath) → nothing
+    Removes a client
+   
+** Signals:
+*** ClientAdded(o: ObjectPath)
+    A new client was added.
+   
+*** ClientNotFound(s: Fingerprint, s: Address)
+    A client connected from Address using Fingerprint, but was
+    rejected because it was not found in the server.  The fingerprint
+    is represented as a string of hexadecimal digits.  The address is
+    an IPv4 or IPv6 address in its normal string format.
+   
+*** ClientRemoved(o: ObjectPath, s: Name)
+    A client named Name on ObjectPath was removed.
+
+
+* Mandos Client Interface:
+  Interface name: "se.bsnet.fukt.Mandos.Client"
+  
+** Methods
+*** Approve(b: Approve) → nothing
+    Approve or deny a connected client waiting for approval.  If
+    denied, a client will not be sent its secret.
+    
+*** CheckedOK() → nothing
+    Assert that this client has been checked and found to be alive.
+    This will restart the timeout before disabling this client.  See
+    also the "LastCheckedOK" property.
+    
+*** Disable() → nothing
+    Disable this client.  See also the "Enabled" property.
+    
+*** Enable() → nothing
+    Enable this client.  See also the "Enabled" property.
+    
+*** StartChecker() → nothing
+    Start a new checker for this client, if none is currently
+    running.  See also the "CheckerRunning" property.
+    
+*** StopChecker() → nothing
+    Abort a running checker process for this client, if any.  See also
+    the "CheckerRunning" property.
+
+** Properties
+   
+   Note: Many of these properties directly correspond to a setting in
+   "clients.conf", in which case they are fully documented in
+   mandos-clients.conf(5).
+   
+   | Name                 | Type | Access     | clients.conf        |
+   |----------------------+------+------------+---------------------|
+   | ApprovedByDefault    | b    | Read/Write | approved_by_default |
+   | ApprovalDelay (a)    | t    | Read/Write | approval_delay      |
+   | ApprovalDuration (a) | t    | Read/Write | approval_duration   |
+   | ApprovalPending (b)  | b    | Read       | N/A                 |
+   | Checker              | s    | Read/Write | checker             |
+   | CheckerRunning (c)   | b    | Read/Write | N/A                 |
+   | Created (d)          | s    | Read       | N/A                 |
+   | Enabled (e)          | b    | Read/Write | N/A                 |
+   | Fingerprint          | s    | Read       | fingerprint         |
+   | Host                 | s    | Read/Write | host                |
+   | Interval (a)         | t    | Read/Write | interval            |
+   | LastCheckedOK (f)    | s    | Read/Write | N/A                 |
+   | LastEnabled (g)      | s    | Read       | N/A                 |
+   | Name                 | s    | Read       | (Section name)      |
+   | ObjectPath           | o    | Read       | N/A                 |
+   | Secret (h)           | ay   | Write      | secret (or secfile) |
+   | Timeout (a)          | t    | Read/Write | timeout             |
+   
+   a) Represented as milliseconds.
+   
+   b) An approval is currently pending.
+   
+   c) Setting this property is equivalent to calling StartChecker() or
+      StopChecker().
+   
+   d) The creation time of this client object, as a RFC 3339 string.
+   
+   e) Setting this property is equivalent to calling Enable() or
+      Disable().
+   
+   f) The last time a checker was successful, as a RFC 3339 string, or
+      an empty string if this has not happened.  Setting this property
+      is equivalent to calling CheckedOK(), i.e. the current time is
+      set, regardless of the string sent.  Please always use an empty
+      string when setting this property, to allow for possible future
+      expansion.
+   
+   g) The last time this client was enabled, as a RFC 3339 string, or
+      an empty string if this has not happened.
+   
+   h) A raw byte array, not hexadecimal digits.
+
+** Signals
+*** CheckerCompleted(n: Exitcode, x: Waitstatus, s: Command)
+    A checker (Command) has completed.  Exitcode is either the exit
+    code or -1 for abnormal exit.  In any case, the full Waitstatus
+    (as from wait(2)) is also available.
+    
+*** CheckerStarted(s: Command)
+    A checker command (Command) has just been started.
+    
+*** GotSecret()
+    This client has been sent its secret.
+    
+*** NeedApproval(t: Timeout, b: ApprovedByDefault)
+    This client will be approved or denied in exactly Timeout
+    milliseconds, depending on ApprovedByDefault.  Approve() can now
+    usefully be called on this client object.
+
+*** PropertyChanged(s: Property, v: Value)
+    The Property on this client has changed to Value.
+
+*** Rejected(s: Reason)
+    This client was not given its secret for a specified Reason.
+
+
+#+STARTUP: showall

=== modified file 'Makefile'
--- Makefile	2009-11-19 18:31:28 +0000
+++ Makefile	2010-09-09 18:28:57 +0000
@@ -209,6 +209,17 @@
 
 # Run the client with a local config and key
 run-client: all keydir/seckey.txt keydir/pubkey.txt
+	@echo "###################################################################"
+	@echo "# The following error messages are harmless and can be safely     #"
+	@echo "# ignored.  The messages are caused by not running as root, but   #"
+	@echo "# you should NOT run \"make run-client\" as root unless you also    #"
+	@echo "# unpacked and compiled Mandos as root, which is NOT recommended. #"
+	@echo "# From plugin-runner: setuid: Operation not permitted             #"
+	@echo "# From askpass-fifo:  mkfifo: Permission denied                   #"
+	@echo "# From mandos-client: setuid: Operation not permitted             #"
+	@echo "#                     seteuid: Operation not permitted            #"
+	@echo "#                     klogctl: Operation not permitted            #"
+	@echo "###################################################################"
 	./plugin-runner --plugin-dir=plugins.d \
 		--config-file=plugin-runner.conf \
 		--options-for=mandos-client:--seckey=keydir/seckey.txt,--pubkey=keydir/pubkey.txt \
@@ -221,6 +232,13 @@
 
 # Run the server with a local config
 run-server: confdir/mandos.conf confdir/clients.conf
+	@echo "#################################################################"
+	@echo "# NOTE: Please IGNORE errors about \"No permission to bind to    #"
+	@echo "# interface\" or \"Could not open file u'/var/run/mandos.pid'\" -  #"
+	@echo "# they are harmless and are caused by the server not running as #"
+	@echo "# root.  Do NOT run \"make run-server\" server as root if you did #"
+	@echo "# not also unpack and compile it as root.                       #"
+	@echo "#################################################################"
 	./mandos --debug --no-dbus --configdir=confdir $(SERVERARGS)
 
 # Used by run-server

=== modified file 'TODO'
--- TODO	2010-04-14 06:36:36 +0000
+++ TODO	2010-09-12 02:51:19 +0000
@@ -1,9 +1,13 @@
 -*- org -*-
 
+* Use _attribute_((nonnull)) wherever possible.
+
+* Release critical
+** TODO Change make-run-server to not include --no-dbus
+
 * mandos-client
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO use error() instead of perror()
 ** TODO [#B] Retry a server which has a non-definite reply:
 *** A closed connection during the TLS handshake
 *** A TCP timeout
@@ -13,31 +17,31 @@
 * splashy
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 
 * usplash
 ** TODO [#A] Make it work again
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
+** TODO Use [[info:libc:Argz%20Functions][argz_extract]]
 
 * askpass-fifo
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] Drop privileges after opening FIFO.
 
 * password-prompt
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] lock stdin (with flock()?)
 
+* plymouth
+** TODO Use [[info:libc:Argz%20Functions][argz_extract]]
+
 * TODO [#B] passdev
 
 * plugin-runner
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#C] use same file name rules as run-parts(8)
 ** kernel command line option for debug info
-** TODO [#B] use error() instead of perror()
+** TODO [#B] Use openat()
 
 * mandos (server)
 ** TODO [#B] Log level							  :BUGS:
@@ -71,19 +75,21 @@
 ** More D-Bus methods
 *** NeedsApproval(50, True) -> timeout, default approve
     Default approval is configurable, but True by default
-    + Approval(True) -> approve sending saved
-    + Approval(False) -> Close client connection immediately
+    + Approve(True) -> approve sending saved
+    + Approve(False) -> Close client connection immediately
 *** NeedsPassword(50) - Timeout, default disapprove
     + SetPass(u"gazonk", True) -> Approval, persistent
-    + Approval(False) -> Close client connection immediately
+    + Approve(False) -> Close client connection immediately
+** TODO [#C] python-parsedatetime
+** TODO [#C] systemd/launchd
+   http://0pointer.de/blog/projects/systemd.html
+** TODO Separate logging logic to own object
+** TODO make clients to a dict!
+** TODO [#A] Limit approval_delay to max gnutls/tls timeout value
 
 * mandos.xml
-** [[file:mandos.xml::XXX][Document D-Bus interface]]
-   Remove mention of lack of such interface in BUGS section
 ** Add mandos contact info in manual pages
 
-* TODO [#A] Provide and install /etc/dbus-1/system.d/mandos.conf
-
 * mandos-ctl
 *** Handle "no D-Bus server" and/or "no Mandos server found" better
 *** [#B] --dump option
@@ -94,9 +100,12 @@
   arguments.
 
 * mandos-monitor
+** TODO help should be toggable
 ** Urwid client data displayer
    Better view of client data in the listing
 *** Properties popup
+** Nicer crashes. Stack traces Messes up shell.
+*** Print a nice "We are sorry" message, save stack trace to log.
 
 * mandos-keygen
 ** TODO Loop until passwords match when run interactively
@@ -106,6 +115,8 @@
    For testing decryption before rebooting.
 
 * Makefile
+** TODO Add "--Xlinker --as-needed"
+   http://udrepper.livejournal.com/19395.html
 ** TODO [#C] Implement DEB_BUILD_OPTIONS
    http://www.debian.org/doc/debian-policy/ch-source.html#s-debianrules-options
 
@@ -117,5 +128,9 @@
 ** TODO [#C] /etc/bash_completion.d/mandos
    From XML sources directly?
 
+* Side Stuff
+** TODO Locate which packet move the other bin/sh when busy box is deactivated
+** TODO contact owner of packet, and ask them to have that shell static in position regardless of busybox
+
 
 #+STARTUP: showall

=== modified file 'clients.conf'
--- clients.conf	2009-12-25 23:13:47 +0000
+++ clients.conf	2010-09-12 03:00:40 +0000
@@ -16,6 +16,15 @@
 # What command to run as "the checker".
 ;checker = fping -q -- %%(host)s
 
+# Whether to approve a client by default after the approval delay.
+;approved_by_default = True
+
+# How long to wait for approval.
+;approval_delay = 0s
+
+# How long one approval will last.
+;approval_duration = 1s
+
 
 ;####
 ;# Example client
@@ -62,4 +71,9 @@
 ;
 ;# Parameters from the [DEFAULT] section can be overridden per client.
 ;interval = 5m
+;
+;# This client requires manual approval before it receives its secret.
+;approved_by_default = False
+;# Require approval within 30 seconds.
+;approval_delay = 30s
 ;####

=== modified file 'debian/control'
--- debian/control	2010-01-01 15:17:12 +0000
+++ debian/control	2010-09-09 22:26:35 +0000
@@ -15,7 +15,8 @@
 Package: mandos
 Architecture: all
 Depends: ${misc:Depends}, python (>=2.5), python-gnutls, python-dbus,
-	python-avahi, python-gobject, avahi-daemon, adduser
+	python-avahi, python-gobject, avahi-daemon, adduser,
+	python-urwid
 Recommends: fping
 Description: a server giving encrypted passwords to Mandos clients
  This is the server part of the Mandos system, which allows

=== modified file 'debian/mandos.docs'
--- debian/mandos.docs	2008-10-18 11:17:22 +0000
+++ debian/mandos.docs	2010-09-12 03:00:40 +0000
@@ -1,3 +1,4 @@
 NEWS
 README
 TODO
+DBUS-API

=== modified file 'mandos'
--- mandos	2010-04-14 06:36:36 +0000
+++ mandos	2010-09-12 02:51:19 +0000
@@ -60,6 +60,7 @@
 import fcntl
 import functools
 import cPickle as pickle
+import multiprocessing
 
 import dbus
 import dbus.service
@@ -82,6 +83,7 @@
 
 version = "1.0.14"
 
+#logger = logging.getLogger(u'mandos')
 logger = logging.Logger(u'mandos')
 syslogger = (logging.handlers.SysLogHandler
              (facility = logging.handlers.SysLogHandler.LOG_DAEMON,
@@ -155,15 +157,20 @@
                             u" after %i retries, exiting.",
                             self.rename_count)
             raise AvahiServiceError(u"Too many renames")
-        self.name = self.server.GetAlternativeServiceName(self.name)
+        self.name = unicode(self.server.GetAlternativeServiceName(self.name))
         logger.info(u"Changing Zeroconf service name to %r ...",
-                    unicode(self.name))
+                    self.name)
         syslogger.setFormatter(logging.Formatter
                                (u'Mandos (%s) [%%(process)d]:'
                                 u' %%(levelname)s: %%(message)s'
                                 % self.name))
         self.remove()
-        self.add()
+        try:
+            self.add()
+        except dbus.exceptions.DBusException, error:
+            logger.critical(u"DBusException: %s", error)
+            self.cleanup()
+            os._exit(1)
         self.rename_count += 1
     def remove(self):
         """Derived from the Avahi example code"""
@@ -192,7 +199,7 @@
         self.group.Commit()
     def entry_group_state_changed(self, state, error):
         """Derived from the Avahi example code"""
-        logger.debug(u"Avahi state change: %i", state)
+        logger.debug(u"Avahi entry group state change: %i", state)
         
         if state == avahi.ENTRY_GROUP_ESTABLISHED:
             logger.debug(u"Zeroconf service established.")
@@ -211,6 +218,7 @@
             self.group = None
     def server_state_changed(self, state):
         """Derived from the Avahi example code"""
+        logger.debug(u"Avahi server state change: %i", state)
         if state == avahi.SERVER_COLLISION:
             logger.error(u"Zeroconf server name collision")
             self.remove()
@@ -257,6 +265,9 @@
                      runtime with vars(self) as dict, so that for
                      instance %(name)s can be used in the command.
     current_checker_command: string; current running checker_command
+    approval_delay: datetime.timedelta(); Time to wait for approval
+    _approved:   bool(); 'None' if not yet approved/disapproved
+    approval_duration: datetime.timedelta(); Duration of one approval
     """
     
     @staticmethod
@@ -273,6 +284,9 @@
     def interval_milliseconds(self):
         "Return the 'interval' attribute in milliseconds"
         return self._timedelta_to_milliseconds(self.interval)
+
+    def approval_delay_milliseconds(self):
+        return self._timedelta_to_milliseconds(self.approval_delay)
     
     def __init__(self, name = None, disable_hook=None, config=None):
         """Note: the 'checker' key in 'config' sets the
@@ -313,12 +327,27 @@
         self.checker_command = config[u"checker"]
         self.current_checker_command = None
         self.last_connect = None
+        self._approved = None
+        self.approved_by_default = config.get(u"approved_by_default",
+                                              True)
+        self.approvals_pending = 0
+        self.approval_delay = string_to_delta(
+            config[u"approval_delay"])
+        self.approval_duration = string_to_delta(
+            config[u"approval_duration"])
+        self.changedstate = multiprocessing_manager.Condition(multiprocessing_manager.Lock())
     
+    def send_changedstate(self):
+        self.changedstate.acquire()
+        self.changedstate.notify_all()
+        self.changedstate.release()
+        
     def enable(self):
         """Start this client's checker and timeout hooks"""
         if getattr(self, u"enabled", False):
             # Already enabled
             return
+        self.send_changedstate()
         self.last_enabled = datetime.datetime.utcnow()
         # Schedule a new checker to be started an 'interval' from now,
         # and every interval from then on.
@@ -338,6 +367,8 @@
         if not getattr(self, "enabled", False):
             return False
         if not quiet:
+            self.send_changedstate()
+        if not quiet:
             logger.info(u"Disabling client %s", self.name)
         if getattr(self, u"disable_initiator_tag", False):
             gobject.source_remove(self.disable_initiator_tag)
@@ -476,7 +507,6 @@
                 raise
         self.checker = None
 
-
 def dbus_service_property(dbus_interface, signature=u"v",
                           access=u"readwrite", byte_arrays=False):
     """Decorators for marking methods of a DBusObjectWithProperties to
@@ -676,6 +706,7 @@
     # dbus.service.Object doesn't use super(), so we can't either.
     
     def __init__(self, bus = None, *args, **kwargs):
+        self._approvals_pending = 0
         self.bus = bus
         Client.__init__(self, *args, **kwargs)
         # Only now, when this client is initialized, can it show up on
@@ -685,6 +716,22 @@
                                   + self.name.replace(u".", u"_")))
         DBusObjectWithProperties.__init__(self, self.bus,
                                           self.dbus_object_path)
+        
+    def _get_approvals_pending(self):
+        return self._approvals_pending
+    def _set_approvals_pending(self, value):
+        old_value = self._approvals_pending
+        self._approvals_pending = value
+        bval = bool(value)
+        if (hasattr(self, "dbus_object_path")
+            and bval is not bool(old_value)):
+            dbus_bool = dbus.Boolean(bval, variant_level=1)
+            self.PropertyChanged(dbus.String(u"ApprovalPending"),
+                                 dbus_bool)
+
+    approvals_pending = property(_get_approvals_pending,
+                                 _set_approvals_pending)
+    del _get_approvals_pending, _set_approvals_pending
     
     @staticmethod
     def _datetime_to_dbus(dt, variant_level=0):
@@ -697,10 +744,10 @@
         r = Client.enable(self)
         if oldstate != self.enabled:
             # Emit D-Bus signals
-            self.PropertyChanged(dbus.String(u"enabled"),
+            self.PropertyChanged(dbus.String(u"Enabled"),
                                  dbus.Boolean(True, variant_level=1))
             self.PropertyChanged(
-                dbus.String(u"last_enabled"),
+                dbus.String(u"LastEnabled"),
                 self._datetime_to_dbus(self.last_enabled,
                                        variant_level=1))
         return r
@@ -710,7 +757,7 @@
         r = Client.disable(self, quiet=quiet)
         if not quiet and oldstate != self.enabled:
             # Emit D-Bus signal
-            self.PropertyChanged(dbus.String(u"enabled"),
+            self.PropertyChanged(dbus.String(u"Enabled"),
                                  dbus.Boolean(False, variant_level=1))
         return r
     
@@ -728,7 +775,7 @@
         self.checker_callback_tag = None
         self.checker = None
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"checker_running"),
+        self.PropertyChanged(dbus.String(u"CheckerRunning"),
                              dbus.Boolean(False, variant_level=1))
         if os.WIFEXITED(condition):
             exitstatus = os.WEXITSTATUS(condition)
@@ -749,7 +796,7 @@
         r = Client.checked_ok(self, *args, **kwargs)
         # Emit D-Bus signal
         self.PropertyChanged(
-            dbus.String(u"last_checked_ok"),
+            dbus.String(u"LastCheckedOK"),
             (self._datetime_to_dbus(self.last_checked_ok,
                                     variant_level=1)))
         return r
@@ -767,7 +814,7 @@
             # Emit D-Bus signal
             self.CheckerStarted(self.current_checker_command)
             self.PropertyChanged(
-                dbus.String(u"checker_running"),
+                dbus.String(u"CheckerRunning"),
                 dbus.Boolean(True, variant_level=1))
         return r
     
@@ -776,9 +823,21 @@
         r = Client.stop_checker(self, *args, **kwargs)
         if (old_checker is not None
             and getattr(self, u"checker", None) is None):
-            self.PropertyChanged(dbus.String(u"checker_running"),
+            self.PropertyChanged(dbus.String(u"CheckerRunning"),
                                  dbus.Boolean(False, variant_level=1))
         return r
+
+    def _reset_approved(self):
+        self._approved = None
+        return False
+    
+    def approve(self, value=True):
+        self.send_changedstate()
+        self._approved = value
+        gobject.timeout_add(self._timedelta_to_milliseconds
+                            (self.approval_duration),
+                            self._reset_approved)
+    
     
     ## D-Bus methods, signals & properties
     _interface = u"se.bsnet.fukt.Mandos.Client"
@@ -806,17 +865,31 @@
     # GotSecret - signal
     @dbus.service.signal(_interface)
     def GotSecret(self):
-        "D-Bus signal"
+        """D-Bus signal
+        Is sent after a successful transfer of secret from the Mandos
+        server to mandos-client
+        """
         pass
     
     # Rejected - signal
-    @dbus.service.signal(_interface)
-    def Rejected(self):
+    @dbus.service.signal(_interface, signature=u"s")
+    def Rejected(self, reason):
+        "D-Bus signal"
+        pass
+    
+    # NeedApproval - signal
+    @dbus.service.signal(_interface, signature=u"tb")
+    def NeedApproval(self, timeout, default):
         "D-Bus signal"
         pass
     
     ## Methods
-    
+
+    # Approve - method
+    @dbus.service.method(_interface, in_signature=u"b")
+    def Approve(self, value):
+        self.approve(value)
+
     # CheckedOK - method
     @dbus.service.method(_interface)
     def CheckedOK(self):
@@ -847,43 +920,82 @@
     
     ## Properties
     
-    # name - property
+    # ApprovalPending - property
+    @dbus_service_property(_interface, signature=u"b", access=u"read")
+    def ApprovalPending_dbus_property(self):
+        return dbus.Boolean(bool(self.approvals_pending))
+    
+    # ApprovedByDefault - property
+    @dbus_service_property(_interface, signature=u"b",
+                           access=u"readwrite")
+    def ApprovedByDefault_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.Boolean(self.approved_by_default)
+        self.approved_by_default = bool(value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovedByDefault"),
+                             dbus.Boolean(value, variant_level=1))
+    
+    # ApprovalDelay - property
+    @dbus_service_property(_interface, signature=u"t",
+                           access=u"readwrite")
+    def ApprovalDelay_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.UInt64(self.approval_delay_milliseconds())
+        self.approval_delay = datetime.timedelta(0, 0, 0, value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovalDelay"),
+                             dbus.UInt64(value, variant_level=1))
+    
+    # ApprovalDuration - property
+    @dbus_service_property(_interface, signature=u"t",
+                           access=u"readwrite")
+    def ApprovalDuration_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.UInt64(self._timedelta_to_milliseconds(
+                    self.approval_duration))
+        self.approval_duration = datetime.timedelta(0, 0, 0, value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovalDuration"),
+                             dbus.UInt64(value, variant_level=1))
+    
+    # Name - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def name_dbus_property(self):
+    def Name_dbus_property(self):
         return dbus.String(self.name)
     
-    # fingerprint - property
+    # Fingerprint - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def fingerprint_dbus_property(self):
+    def Fingerprint_dbus_property(self):
         return dbus.String(self.fingerprint)
     
-    # host - property
+    # Host - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def host_dbus_property(self, value=None):
+    def Host_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.String(self.host)
         self.host = value
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"host"),
+        self.PropertyChanged(dbus.String(u"Host"),
                              dbus.String(value, variant_level=1))
     
-    # created - property
+    # Created - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def created_dbus_property(self):
+    def Created_dbus_property(self):
         return dbus.String(self._datetime_to_dbus(self.created))
     
-    # last_enabled - property
+    # LastEnabled - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def last_enabled_dbus_property(self):
+    def LastEnabled_dbus_property(self):
         if self.last_enabled is None:
             return dbus.String(u"")
         return dbus.String(self._datetime_to_dbus(self.last_enabled))
     
-    # enabled - property
+    # Enabled - property
     @dbus_service_property(_interface, signature=u"b",
                            access=u"readwrite")
-    def enabled_dbus_property(self, value=None):
+    def Enabled_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.Boolean(self.enabled)
         if value:
@@ -891,10 +1003,10 @@
         else:
             self.disable()
     
-    # last_checked_ok - property
+    # LastCheckedOK - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def last_checked_ok_dbus_property(self, value=None):
+    def LastCheckedOK_dbus_property(self, value=None):
         if value is not None:
             self.checked_ok()
             return
@@ -903,15 +1015,15 @@
         return dbus.String(self._datetime_to_dbus(self
                                                   .last_checked_ok))
     
-    # timeout - property
+    # Timeout - property
     @dbus_service_property(_interface, signature=u"t",
                            access=u"readwrite")
-    def timeout_dbus_property(self, value=None):
+    def Timeout_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.UInt64(self.timeout_milliseconds())
         self.timeout = datetime.timedelta(0, 0, 0, value)
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"timeout"),
+        self.PropertyChanged(dbus.String(u"Timeout"),
                              dbus.UInt64(value, variant_level=1))
         if getattr(self, u"disable_initiator_tag", None) is None:
             return
@@ -931,15 +1043,15 @@
             self.disable_initiator_tag = (gobject.timeout_add
                                           (time_to_die, self.disable))
     
-    # interval - property
+    # Interval - property
     @dbus_service_property(_interface, signature=u"t",
                            access=u"readwrite")
-    def interval_dbus_property(self, value=None):
+    def Interval_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.UInt64(self.interval_milliseconds())
         self.interval = datetime.timedelta(0, 0, 0, value)
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"interval"),
+        self.PropertyChanged(dbus.String(u"Interval"),
                              dbus.UInt64(value, variant_level=1))
         if getattr(self, u"checker_initiator_tag", None) is None:
             return
@@ -949,22 +1061,22 @@
                                       (value, self.start_checker))
         self.start_checker()    # Start one now, too
 
-    # checker - property
+    # Checker - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def checker_dbus_property(self, value=None):
+    def Checker_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.String(self.checker_command)
         self.checker_command = value
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"checker"),
+        self.PropertyChanged(dbus.String(u"Checker"),
                              dbus.String(self.checker_command,
                                          variant_level=1))
     
-    # checker_running - property
+    # CheckerRunning - property
     @dbus_service_property(_interface, signature=u"b",
                            access=u"readwrite")
-    def checker_running_dbus_property(self, value=None):
+    def CheckerRunning_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.Boolean(self.checker is not None)
         if value:
@@ -972,20 +1084,46 @@
         else:
             self.stop_checker()
     
-    # object_path - property
+    # ObjectPath - property
     @dbus_service_property(_interface, signature=u"o", access=u"read")
-    def object_path_dbus_property(self):
+    def ObjectPath_dbus_property(self):
         return self.dbus_object_path # is already a dbus.ObjectPath
     
-    # secret = property
+    # Secret = property
     @dbus_service_property(_interface, signature=u"ay",
                            access=u"write", byte_arrays=True)
-    def secret_dbus_property(self, value):
+    def Secret_dbus_property(self, value):
         self.secret = str(value)
     
     del _interface
 
 
+class ProxyClient(object):
+    def __init__(self, child_pipe, fpr, address):
+        self._pipe = child_pipe
+        self._pipe.send(('init', fpr, address))
+        if not self._pipe.recv():
+            raise KeyError()
+
+    def __getattribute__(self, name):
+        if(name == '_pipe'):
+            return super(ProxyClient, self).__getattribute__(name)
+        self._pipe.send(('getattr', name))
+        data = self._pipe.recv()
+        if data[0] == 'data':
+            return data[1]
+        if data[0] == 'function':
+            def func(*args, **kwargs):
+                self._pipe.send(('funcall', name, args, kwargs))
+                return self._pipe.recv()[1]
+            return func
+
+    def __setattr__(self, name, value):
+        if(name == '_pipe'):
+            return super(ProxyClient, self).__setattr__(name, value)
+        self._pipe.send(('setattr', name, value))
+
+
 class ClientHandler(socketserver.BaseRequestHandler, object):
     """A class to handle client connections.
     
@@ -993,33 +1131,22 @@
     Note: This will run in its own forked process."""
     
     def handle(self):
-        logger.info(u"TCP connection from: %s",
-                    unicode(self.client_address))
-        logger.debug(u"IPC Pipe FD: %d",
-                     self.server.child_pipe[1].fileno())
-        # Open IPC pipe to parent process
-        with contextlib.nested(self.server.child_pipe[1],
-                               self.server.parent_pipe[0]
-                               ) as (ipc, ipc_return):
+        with contextlib.closing(self.server.child_pipe) as child_pipe:
+            logger.info(u"TCP connection from: %s",
+                        unicode(self.client_address))
+            logger.debug(u"Pipe FD: %d",
+                         self.server.child_pipe.fileno())
+
             session = (gnutls.connection
                        .ClientSession(self.request,
                                       gnutls.connection
                                       .X509Credentials()))
-            
-            line = self.request.makefile().readline()
-            logger.debug(u"Protocol version: %r", line)
-            try:
-                if int(line.strip().split()[0]) > 1:
-                    raise RuntimeError
-            except (ValueError, IndexError, RuntimeError), error:
-                logger.error(u"Unknown protocol version: %s", error)
-                return
-            
+
             # Note: gnutls.connection.X509Credentials is really a
             # generic GnuTLS certificate credentials object so long as
             # no X.509 keys are added to it.  Therefore, we can use it
             # here despite using OpenPGP certificates.
-            
+
             #priority = u':'.join((u"NONE", u"+VERS-TLS1.1",
             #                      u"+AES-256-CBC", u"+SHA1",
             #                      u"+COMP-NULL", u"+CTYPE-OPENPGP",
@@ -1031,7 +1158,19 @@
             (gnutls.library.functions
              .gnutls_priority_set_direct(session._c_object,
                                          priority, None))
-            
+
+            # Start communication using the Mandos protocol
+            # Get protocol number
+            line = self.request.makefile().readline()
+            logger.debug(u"Protocol version: %r", line)
+            try:
+                if int(line.strip().split()[0]) > 1:
+                    raise RuntimeError
+            except (ValueError, IndexError, RuntimeError), error:
+                logger.error(u"Unknown protocol version: %s", error)
+                return
+
+            # Start GnuTLS connection
             try:
                 session.handshake()
             except gnutls.errors.GNUTLSError, error:
@@ -1040,6 +1179,8 @@
                 # established.  Just abandon the request.
                 return
             logger.debug(u"Handshake succeeded")
+
+            approval_required = False
             try:
                 try:
                     fpr = self.fingerprint(self.peer_certificate
@@ -1049,34 +1190,92 @@
                     return
                 logger.debug(u"Fingerprint: %s", fpr)
 
-                for c in self.server.clients:
-                    if c.fingerprint == fpr:
-                        client = c
+                try:
+                    client = ProxyClient(child_pipe, fpr,
+                                         self.client_address)
+                except KeyError:
+                    return
+                
+                if client.approval_delay:
+                    delay = client.approval_delay
+                    client.approvals_pending += 1
+                    approval_required = True
+                
+                while True:
+                    if not client.enabled:
+                        logger.warning(u"Client %s is disabled",
+                                       client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.Rejected("Disabled")                    
+                        return
+                    
+                    if client._approved or not client.approval_delay:
+                        #We are approved or approval is disabled
                         break
-                else:
-                    ipc.write(u"NOTFOUND %s %s\n"
-                              % (fpr, unicode(self.client_address)))
-                    return
-                # Have to check if client.enabled, since it is
-                # possible that the client was disabled since the
-                # GnuTLS session was established.
-                ipc.write(u"GETATTR enabled %s\n" % fpr)
-                enabled = pickle.load(ipc_return)
-                if not enabled:
-                    ipc.write(u"DISABLED %s\n" % client.name)
-                    return
-                # Send "NEED_APPROVAL" here and hang waiting
-                # for response?  Leave timeout to parent process?
-                ipc.write(u"SENDING %s\n" % client.name)
+                    elif client._approved is None:
+                        logger.info(u"Client %s needs approval",
+                                    client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.NeedApproval(
+                                client.approval_delay_milliseconds(),
+                                client.approved_by_default)
+                    else:
+                        logger.warning(u"Client %s was not approved",
+                                       client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.Rejected("Denied")
+                        return
+                    
+                    #wait until timeout or approved
+                    #x = float(client._timedelta_to_milliseconds(delay))
+                    time = datetime.datetime.now()
+                    client.changedstate.acquire()
+                    client.changedstate.wait(float(client._timedelta_to_milliseconds(delay) / 1000))
+                    client.changedstate.release()
+                    time2 = datetime.datetime.now()
+                    if (time2 - time) >= delay:
+                        if not client.approved_by_default:
+                            logger.warning("Client %s timed out while"
+                                           " waiting for approval",
+                                           client.name)
+                            if self.server.use_dbus:
+                                # Emit D-Bus signal
+                                client.Rejected("Timed out")
+                            return
+                        else:
+                            break
+                    else:
+                        delay -= time2 - time
+                
                 sent_size = 0
                 while sent_size < len(client.secret):
-                    sent = session.send(client.secret[sent_size:])
+                    try:
+                        sent = session.send(client.secret[sent_size:])
+                    except (gnutls.errors.GNUTLSError), error:
+                        logger.warning("gnutls send failed")
+                        return
                     logger.debug(u"Sent: %d, remaining: %d",
                                  sent, len(client.secret)
                                  - (sent_size + sent))
                     sent_size += sent
+
+                logger.info(u"Sending secret to %s", client.name)
+                # bump the timeout as if seen
+                client.checked_ok()
+                if self.server.use_dbus:
+                    # Emit D-Bus signal
+                    client.GotSecret()
+            
             finally:
-                session.bye()
+                if approval_required:
+                    client.approvals_pending -= 1
+                try:
+                    session.bye()
+                except (gnutls.errors.GNUTLSError), error:
+                    logger.warning("GnuTLS bye failed")
     
     @staticmethod
     def peer_certificate(session):
@@ -1142,30 +1341,39 @@
         return hex_fpr
 
 
-class ForkingMixInWithPipes(socketserver.ForkingMixIn, object):
-    """Like socketserver.ForkingMixIn, but also pass a pipe pair."""
+class MultiprocessingMixIn(object):
+    """Like socketserver.ThreadingMixIn, but with multiprocessing"""
+    def sub_process_main(self, request, address):
+        try:
+            self.finish_request(request, address)
+        except:
+            self.handle_error(request, address)
+        self.close_request(request)
+            
+    def process_request(self, request, address):
+        """Start a new process to process the request."""
+        multiprocessing.Process(target = self.sub_process_main,
+                                args = (request, address)).start()
+
+class MultiprocessingMixInWithPipe(MultiprocessingMixIn, object):
+    """ adds a pipe to the MixIn """
     def process_request(self, request, client_address):
         """Overrides and wraps the original process_request().
         
         This function creates a new pipe in self.pipe
         """
-        # Child writes to child_pipe
-        self.child_pipe = map(os.fdopen, os.pipe(), u"rw", (1, 0))
-        # Parent writes to parent_pipe
-        self.parent_pipe = map(os.fdopen, os.pipe(), u"rw", (1, 0))
-        super(ForkingMixInWithPipes,
+        parent_pipe, self.child_pipe = multiprocessing.Pipe()
+
+        super(MultiprocessingMixInWithPipe,
               self).process_request(request, client_address)
-        # Close unused ends for parent
-        self.parent_pipe[0].close() # close read end
-        self.child_pipe[1].close()  # close write end
-        self.add_pipe_fds(self.child_pipe[0], self.parent_pipe[1])
-    def add_pipe_fds(self, child_pipe_fd, parent_pipe_fd):
+        self.child_pipe.close()
+        self.add_pipe(parent_pipe)
+
+    def add_pipe(self, parent_pipe):
         """Dummy function; override as necessary"""
-        child_pipe_fd.close()
-        parent_pipe_fd.close()
-
-
-class IPv6_TCPServer(ForkingMixInWithPipes,
+        pass
+
+class IPv6_TCPServer(MultiprocessingMixInWithPipe,
                      socketserver.TCPServer, object):
     """IPv6-capable TCP server.  Accepts 'None' as address and/or port
     
@@ -1256,14 +1464,15 @@
             return socketserver.TCPServer.server_activate(self)
     def enable(self):
         self.enabled = True
-    def add_pipe_fds(self, child_pipe_fd, parent_pipe_fd):
+    def add_pipe(self, parent_pipe):
         # Call "handle_ipc" for both data and EOF events
-        gobject.io_add_watch(child_pipe_fd.fileno(),
+        gobject.io_add_watch(parent_pipe.fileno(),
                              gobject.IO_IN | gobject.IO_HUP,
                              functools.partial(self.handle_ipc,
-                                               reply = parent_pipe_fd,
-                                               sender= child_pipe_fd))
-    def handle_ipc(self, source, condition, reply=None, sender=None):
+                                               parent_pipe = parent_pipe))
+        
+    def handle_ipc(self, source, condition, parent_pipe=None,
+                   client_object=None):
         condition_names = {
             gobject.IO_IN: u"IN",   # There is data to read.
             gobject.IO_OUT: u"OUT", # Data can be written (without
@@ -1280,68 +1489,60 @@
                                        if cond & condition)
         logger.debug(u"Handling IPC: FD = %d, condition = %s", source,
                      conditions_string)
-        
-        # Read a line from the file object
-        cmdline = sender.readline()
-        if not cmdline:             # Empty line means end of file
-            # close the IPC pipes
-            sender.close()
-            reply.close()
-            
-            # Stop calling this function
-            return False
-        
-        logger.debug(u"IPC command: %r", cmdline)
-        
-        # Parse and act on command
-        cmd, args = cmdline.rstrip(u"\r\n").split(None, 1)
-        
-        if cmd == u"NOTFOUND":
-            fpr, address = args.split(None, 1)
-            logger.warning(u"Client not found for fingerprint: %s, ad"
-                           u"dress: %s", fpr, address)
-            if self.use_dbus:
-                # Emit D-Bus signal
-                mandos_dbus_service.ClientNotFound(fpr, address)
-        elif cmd == u"DISABLED":
-            for client in self.clients:
-                if client.name == args:
-                    logger.warning(u"Client %s is disabled", args)
-                    if self.use_dbus:
-                        # Emit D-Bus signal
-                        client.Rejected()
-                    break
-            else:
-                logger.error(u"Unknown client %s is disabled", args)
-        elif cmd == u"SENDING":
-            for client in self.clients:
-                if client.name == args:
-                    logger.info(u"Sending secret to %s", client.name)
-                    client.checked_ok()
-                    if self.use_dbus:
-                        # Emit D-Bus signal
-                        client.GotSecret()
-                    break
-            else:
-                logger.error(u"Sending secret to unknown client %s",
-                             args)
-        elif cmd == u"GETATTR":
-            attr_name, fpr = args.split(None, 1)
-            for client in self.clients:
-                if client.fingerprint == fpr:
-                    attr_value = getattr(client, attr_name, None)
-                    logger.debug("IPC reply: %r", attr_value)
-                    pickle.dump(attr_value, reply)
-                    break
-            else:
-                logger.error(u"Client %s on address %s requesting "
-                             u"attribute %s not found", fpr, address,
-                             attr_name)
-                pickle.dump(None, reply)
-        else:
-            logger.error(u"Unknown IPC command: %r", cmdline)
-        
-        # Keep calling this function
+
+        # error or the other end of multiprocessing.Pipe has closed
+        if condition & (gobject.IO_ERR | condition & gobject.IO_HUP):
+            return False
+        
+        # Read a request from the child
+        request = parent_pipe.recv()
+        logger.debug(u"IPC request: %s", repr(request))
+        command = request[0]
+        
+        if command == 'init':
+            fpr = request[1]
+            address = request[2]
+            
+            for c in self.clients:
+                if c.fingerprint == fpr:
+                    client = c
+                    break
+            else:
+                logger.warning(u"Client not found for fingerprint: %s, ad"
+                               u"dress: %s", fpr, address)
+                if self.use_dbus:
+                    # Emit D-Bus signal
+                    mandos_dbus_service.ClientNotFound(fpr, address)
+                parent_pipe.send(False)
+                return False
+            
+            gobject.io_add_watch(parent_pipe.fileno(),
+                                 gobject.IO_IN | gobject.IO_HUP,
+                                 functools.partial(self.handle_ipc,
+                                                   parent_pipe = parent_pipe,
+                                                   client_object = client))
+            parent_pipe.send(True)
+            # remove the old hook in favor of the new above hook on same fileno
+            return False
+        if command == 'funcall':
+            funcname = request[1]
+            args = request[2]
+            kwargs = request[3]
+            
+            parent_pipe.send(('data', getattr(client_object, funcname)(*args, **kwargs)))
+
+        if command == 'getattr':
+            attrname = request[1]
+            if callable(client_object.__getattribute__(attrname)):
+                parent_pipe.send(('function',))
+            else:
+                parent_pipe.send(('data', client_object.__getattribute__(attrname)))
+        
+        if command == 'setattr':
+            attrname = request[1]
+            value = request[2]
+            setattr(client_object, attrname, value)
+
         return True
 
 
@@ -1450,6 +1651,8 @@
     parser.add_option("--debug", action=u"store_true",
                       help=u"Debug mode; run in foreground and log to"
                       u" terminal")
+    parser.add_option("--debuglevel", type=u"string", metavar="Level",
+                      help=u"Debug level for stdout output")
     parser.add_option("--priority", type=u"string", help=u"GnuTLS"
                       u" priority string (see GnuTLS documentation)")
     parser.add_option("--servicename", type=u"string",
@@ -1480,6 +1683,7 @@
                         u"servicename": u"Mandos",
                         u"use_dbus": u"True",
                         u"use_ipv6": u"True",
+                        u"debuglevel": u"",
                         }
     
     # Parse config file for server-global settings
@@ -1502,7 +1706,7 @@
     # options, if set.
     for option in (u"interface", u"address", u"port", u"debug",
                    u"priority", u"servicename", u"configdir",
-                   u"use_dbus", u"use_ipv6"):
+                   u"use_dbus", u"use_ipv6", u"debuglevel"):
         value = getattr(options, option)
         if value is not None:
             server_settings[option] = value
@@ -1517,13 +1721,10 @@
     
     # For convenience
     debug = server_settings[u"debug"]
+    debuglevel = server_settings[u"debuglevel"]
     use_dbus = server_settings[u"use_dbus"]
     use_ipv6 = server_settings[u"use_ipv6"]
-    
-    if not debug:
-        syslogger.setLevel(logging.WARNING)
-        console.setLevel(logging.WARNING)
-    
+
     if server_settings[u"servicename"] != u"Mandos":
         syslogger.setFormatter(logging.Formatter
                                (u'Mandos (%s) [%%(process)d]:'
@@ -1535,6 +1736,8 @@
                         u"interval": u"5m",
                         u"checker": u"fping -q -- %%(host)s",
                         u"host": u"",
+                        u"approval_delay": u"0s",
+                        u"approval_duration": u"1s",
                         }
     client_config = configparser.SafeConfigParser(client_defaults)
     client_config.read(os.path.join(server_settings[u"configdir"],
@@ -1546,7 +1749,8 @@
     tcp_server = MandosServer((server_settings[u"address"],
                                server_settings[u"port"]),
                               ClientHandler,
-                              interface=server_settings[u"interface"],
+                              interface=(server_settings[u"interface"]
+                                         or None),
                               use_ipv6=use_ipv6,
                               gnutls_priority=
                               server_settings[u"priority"],
@@ -1578,8 +1782,17 @@
         if error[0] != errno.EPERM:
             raise error
     
-    # Enable all possible GnuTLS debugging
+    if not debug and not debuglevel:
+        syslogger.setLevel(logging.WARNING)
+        console.setLevel(logging.WARNING)
+    if debuglevel:
+        level = getattr(logging, debuglevel.upper())
+        syslogger.setLevel(level)
+        console.setLevel(level)
+
     if debug:
+        # Enable all possible GnuTLS debugging
+        
         # "Use a log level over 10 to enable all debugging options."
         # - GnuTLS manual
         gnutls.library.functions.gnutls_global_set_log_level(11)
@@ -1590,6 +1803,16 @@
         
         (gnutls.library.functions
          .gnutls_global_set_log_function(debug_gnutls))
+        
+        # Redirect stdin so all checkers get /dev/null
+        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
+        os.dup2(null, sys.stdin.fileno())
+        if null > 2:
+            os.close(null)
+    else:
+        # No console logging
+        logger.removeHandler(console)
+    
     
     global main_loop
     # From the Avahi example code
@@ -1613,29 +1836,37 @@
     if server_settings["interface"]:
         service.interface = (if_nametoindex
                              (str(server_settings[u"interface"])))
+
+    if not debug:
+        # Close all input and output, do double fork, etc.
+        daemon()
+        
+    global multiprocessing_manager
+    multiprocessing_manager = multiprocessing.Manager()
     
     client_class = Client
     if use_dbus:
         client_class = functools.partial(ClientDBus, bus = bus)
+    def client_config_items(config, section):
+        special_settings = {
+            "approved_by_default":
+                lambda: config.getboolean(section,
+                                          "approved_by_default"),
+            }
+        for name, value in config.items(section):
+            try:
+                yield (name, special_settings[name]())
+            except KeyError:
+                yield (name, value)
+    
     tcp_server.clients.update(set(
             client_class(name = section,
-                         config= dict(client_config.items(section)))
+                         config= dict(client_config_items(
+                        client_config, section)))
             for section in client_config.sections()))
     if not tcp_server.clients:
         logger.warning(u"No clients defined")
-    
-    if debug:
-        # Redirect stdin so all checkers get /dev/null
-        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
-        os.dup2(null, sys.stdin.fileno())
-        if null > 2:
-            os.close(null)
-    else:
-        # No console logging
-        logger.removeHandler(console)
-        # Close all input and output, do double fork, etc.
-        daemon()
-    
+        
     try:
         with pidfile:
             pid = os.getpid()

=== modified file 'mandos-clients.conf.xml'
--- mandos-clients.conf.xml	2009-12-25 23:13:47 +0000
+++ mandos-clients.conf.xml	2010-09-12 03:00:40 +0000
@@ -3,7 +3,7 @@
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY CONFNAME "mandos-clients.conf">
 <!ENTITY CONFPATH "<filename>/etc/mandos/clients.conf</filename>">
-<!ENTITY TIMESTAMP "2009-12-09">
+<!ENTITY TIMESTAMP "2010-09-12">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -266,6 +266,58 @@
 	</listitem>
       </varlistentry>
       
+      <varlistentry>
+	<term><option>approved_by_default<literal> = </literal
+          >{ <literal >1</literal> | <literal>yes</literal> | <literal
+	  >true</literal> | <literal>on</literal> | <literal
+	  >0</literal> | <literal>no</literal> | <literal
+	  >false</literal> | <literal>off</literal> }</option></term>
+	<listitem>
+	  <para>
+	    Whether to approve a client by default after
+	    the <option>approval_delay</option>.  The default
+	    is <quote>True</quote>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>approval_delay<literal> = </literal><replaceable
+	>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>.
+	  </para>
+	  <para>
+	    How long to wait for external approval before resorting to
+	    use the <option>approved_by_default</option> value.  The
+	    default is <quote>0s</quote>, i.e. not to wait.
+	  </para>
+	  <para>
+	    The format of <replaceable>TIME</replaceable> is the same
+	    as for <varname>timeout</varname> above.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>approval_duration<literal> = </literal
+        ><replaceable>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>.
+	  </para>
+	  <para>
+	    How long an external approval lasts.  The default is 1
+	    second.
+	  </para>
+	  <para>
+	    The format of <replaceable>TIME</replaceable> is the same
+	    as for <varname>timeout</varname> above.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
     </variablelist>
   </refsect1>
   
@@ -379,6 +431,8 @@
 fingerprint = 3e393aeaefb84c7e89e2f547b3a107558fca3a27
 secfile = /etc/mandos/bar-secret
 timeout = 15m
+approved_by_default = False
+approval_delay = 30s
       </programlisting>
     </informalexample>
   </refsect1>

=== modified file 'mandos-ctl'
--- mandos-ctl	2009-12-26 01:19:47 +0000
+++ mandos-ctl	2010-09-12 02:51:19 +0000
@@ -12,27 +12,31 @@
 locale.setlocale(locale.LC_ALL, u'')
 
 tablewords = {
-    'name': u'Name',
-    'enabled': u'Enabled',
-    'timeout': u'Timeout',
-    'last_checked_ok': u'Last Successful Check',
-    'created': u'Created',
-    'interval': u'Interval',
-    'host': u'Host',
-    'fingerprint': u'Fingerprint',
-    'checker_running': u'Check Is Running',
-    'last_enabled': u'Last Enabled',
-    'checker': u'Checker',
+    'Name': u'Name',
+    'Enabled': u'Enabled',
+    'Timeout': u'Timeout',
+    'LastCheckedOK': u'Last Successful Check',
+    'Created': u'Created',
+    'Interval': u'Interval',
+    'Host': u'Host',
+    'Fingerprint': u'Fingerprint',
+    'CheckerRunning': u'Check Is Running',
+    'LastEnabled': u'Last Enabled',
+    'Checker': u'Checker',
     }
-defaultkeywords = ('name', 'enabled', 'timeout', 'last_checked_ok')
+defaultkeywords = ('Name', 'Enabled', 'Timeout', 'LastCheckedOK')
 domain = 'se.bsnet.fukt'
 busname = domain + '.Mandos'
 server_path = '/'
 server_interface = domain + '.Mandos'
 client_interface = domain + '.Mandos.Client'
 version = "1.0.14"
-bus = dbus.SystemBus()
-mandos_dbus_objc = bus.get_object(busname, server_path)
+try:
+    bus = dbus.SystemBus()
+    mandos_dbus_objc = bus.get_object(busname, server_path)
+except dbus.exceptions.DBusException:
+    sys.exit(1)
+    
 mandos_serv = dbus.Interface(mandos_dbus_objc,
                              dbus_interface = server_interface)
 mandos_clients = mandos_serv.GetAllClientsWithProperties()
@@ -104,7 +108,8 @@
     # Create format string to print table rows
     format_string = u' '.join(u'%%-%ds' %
                               max(len(tablewords[key]),
-                                  max(len(valuetostring(client[key], key))
+                                  max(len(valuetostring(client[key],
+                                                        key))
                                       for client in
                                       clients))
                               for key in keywords)
@@ -141,6 +146,10 @@
                   help="Set host for client")
 parser.add_option("-s", "--secret", type="string",
                   help="Set password blob (file) for client")
+parser.add_option("-A", "--approve", action="store_true",
+                  help="Approve any current client request")
+parser.add_option("-D", "--deny", action="store_true",
+                  help="Deny any current client request")
 options, client_names = parser.parse_args()
 
 # Compile list of clients to process
@@ -158,9 +167,9 @@
 if not clients and mandos_clients.values():
     keywords = defaultkeywords
     if options.all:
-        keywords = ('name', 'enabled', 'timeout', 'last_checked_ok',
-                    'created', 'interval', 'host', 'fingerprint',
-                    'checker_running', 'last_enabled', 'checker')
+        keywords = ('Name', 'Enabled', 'Timeout', 'LastCheckedOK',
+                    'Created', 'Interval', 'Host', 'Fingerprint',
+                    'CheckerRunning', 'LastEnabled', 'Checker')
     print_clients(mandos_clients.values())
 
 # Process each client in the list by all selected options
@@ -179,26 +188,32 @@
         client.StopChecker(dbus_interface=client_interface)
     if options.is_enabled:
         sys.exit(0 if client.Get(client_interface,
-                                 u"enabled",
+                                 u"Enabled",
                                  dbus_interface=dbus.PROPERTIES_IFACE)
                  else 1)
     if options.checker:
-        client.Set(client_interface, u"checker", options.checker,
+        client.Set(client_interface, u"Checker", options.checker,
                    dbus_interface=dbus.PROPERTIES_IFACE)
     if options.host:
-        client.Set(client_interface, u"host", options.host,
+        client.Set(client_interface, u"Host", options.host,
                    dbus_interface=dbus.PROPERTIES_IFACE)
     if options.interval:
-        client.Set(client_interface, u"interval",
+        client.Set(client_interface, u"Interval",
                    timedelta_to_milliseconds
                    (string_to_delta(options.interval)),
                    dbus_interface=dbus.PROPERTIES_IFACE)
     if options.timeout:
-        client.Set(client_interface, u"timeout",
+        client.Set(client_interface, u"Timeout",
                    timedelta_to_milliseconds(string_to_delta
                                              (options.timeout)),
                    dbus_interface=dbus.PROPERTIES_IFACE)
     if options.secret:
-        client.Set(client_interface, u"secret",
+        client.Set(client_interface, u"Secret",
                    dbus.ByteArray(open(options.secret, u'rb').read()),
                    dbus_interface=dbus.PROPERTIES_IFACE)
+    if options.approve:
+        client.Approve(dbus.Boolean(True),
+                       dbus_interface=client_interface)
+    if options.deny:
+        client.Approve(dbus.Boolean(False),
+                       dbus_interface=client_interface)

=== modified file 'mandos-monitor'
--- mandos-monitor	2009-12-25 23:13:47 +0000
+++ mandos-monitor	2010-09-12 03:08:51 +0000
@@ -23,11 +23,14 @@
 
 locale.setlocale(locale.LC_ALL, u'')
 
+import logging
+logging.getLogger('dbus.proxies').setLevel(logging.CRITICAL)
+
 # Some useful constants
 domain = 'se.bsnet.fukt'
 server_interface = domain + '.Mandos'
 client_interface = domain + '.Mandos.Client'
-version = "1.0.14"
+version = "1.0.15"
 
 # Always run in monochrome mode
 urwid.curses_display.curses.has_colors = lambda : False
@@ -37,6 +40,22 @@
 urwid.curses_display.curses.A_UNDERLINE |= (
     urwid.curses_display.curses.A_BLINK)
 
+def isoformat_to_datetime(iso):
+    "Parse an ISO 8601 date string to a datetime.datetime()"
+    if not iso:
+        return None
+    d, t = iso.split(u"T", 1)
+    year, month, day = d.split(u"-", 2)
+    hour, minute, second = t.split(u":", 2)
+    second, fraction = divmod(float(second), 1)
+    return datetime.datetime(int(year),
+                             int(month),
+                             int(day),
+                             int(hour),
+                             int(minute),
+                             int(second),           # Whole seconds
+                             int(fraction*1000000)) # Microseconds
+
 class MandosClientPropertyCache(object):
     """This wraps a Mandos Client D-Bus proxy object, caches the
     properties and calls a hook function when any of them are
@@ -50,12 +69,14 @@
                                      self.property_changed,
                                      client_interface,
                                      byte_arrays=True)
-
+        
         self.properties.update(
             self.proxy.GetAll(client_interface,
                               dbus_interface = dbus.PROPERTIES_IFACE))
-        super(MandosClientPropertyCache, self).__init__(
-            proxy_object=proxy_object, *args, **kwargs)
+
+        #XXX This break good super behaviour!
+#        super(MandosClientPropertyCache, self).__init__(
+#            *args, **kwargs)
     
     def property_changed(self, property=None, value=None):
         """This is called whenever we get a PropertyChanged signal
@@ -80,6 +101,9 @@
         # Logger
         self.logger = logger
         
+        self._update_timer_callback_tag = None
+        self.last_checker_failed = False
+        
         # The widget shown normally
         self._text_widget = urwid.Text(u"")
         # The widget shown when we have focus
@@ -101,46 +125,87 @@
                                      self.got_secret,
                                      client_interface,
                                      byte_arrays=True)
+        self.proxy.connect_to_signal(u"NeedApproval",
+                                     self.need_approval,
+                                     client_interface,
+                                     byte_arrays=True)
         self.proxy.connect_to_signal(u"Rejected",
                                      self.rejected,
                                      client_interface,
                                      byte_arrays=True)
+        last_checked_ok = isoformat_to_datetime(self.properties
+                                                [u"LastCheckedOK"])
+        if last_checked_ok is None:
+            self.last_checker_failed = True
+        else:
+            self.last_checker_failed = ((datetime.datetime.utcnow()
+                                         - last_checked_ok)
+                                        > datetime.timedelta
+                                        (milliseconds=
+                                         self.properties
+                                         [u"Interval"]))
+        if self.last_checker_failed:
+            self._update_timer_callback_tag = (gobject.timeout_add
+                                               (1000,
+                                                self.update_timer))
     
     def checker_completed(self, exitstatus, condition, command):
         if exitstatus == 0:
-            self.logger(u'Checker for client %s (command "%s")'
-                        u' was successful'
-                        % (self.properties[u"name"], command))
+            if self.last_checker_failed:
+                self.last_checker_failed = False
+                gobject.source_remove(self._update_timer_callback_tag)
+                self._update_timer_callback_tag = None
+            #self.logger(u'Checker for client %s (command "%s")'
+            #            u' was successful'
+            #            % (self.properties[u"Name"], command))
+            self.update()
             return
+        # Checker failed
+        if not self.last_checker_failed:
+            self.last_checker_failed = True
+            self._update_timer_callback_tag = (gobject.timeout_add
+                                               (1000,
+                                                self.update_timer))
         if os.WIFEXITED(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' failed with exit code %s'
-                        % (self.properties[u"name"], command,
+                        % (self.properties[u"Name"], command,
                            os.WEXITSTATUS(condition)))
-            return
-        if os.WIFSIGNALED(condition):
+        elif os.WIFSIGNALED(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' was killed by signal %s'
-                        % (self.properties[u"name"], command,
+                        % (self.properties[u"Name"], command,
                            os.WTERMSIG(condition)))
-            return
-        if os.WCOREDUMP(condition):
+        elif os.WCOREDUMP(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' dumped core'
-                        % (self.properties[u"name"], command))
-        self.logger(u'Checker for client %s completed mysteriously')
+                        % (self.properties[u"Name"], command))
+        else:
+            self.logger(u'Checker for client %s completed'
+                        u' mysteriously')
+        self.update()
     
     def checker_started(self, command):
-        self.logger(u'Client %s started checker "%s"'
-                    % (self.properties[u"name"], unicode(command)))
+        #self.logger(u'Client %s started checker "%s"'
+        #            % (self.properties[u"Name"], unicode(command)))
+        pass
     
     def got_secret(self):
+        self.last_checker_failed = False
         self.logger(u'Client %s received its secret'
-                    % self.properties[u"name"])
-    
-    def rejected(self):
-        self.logger(u'Client %s was rejected'
-                    % self.properties[u"name"])
+                    % self.properties[u"Name"])
+    
+    def need_approval(self, timeout, default):
+        if not default:
+            message = u'Client %s needs approval within %s seconds'
+        else:
+            message = u'Client %s will get its secret in %s seconds'
+        self.logger(message
+                    % (self.properties[u"Name"], timeout/1000))
+    
+    def rejected(self, reason):
+        self.logger(u'Client %s was rejected; reason: %s'
+                    % (self.properties[u"Name"], reason))
     
     def selectable(self):
         """Make this a "selectable" widget.
@@ -168,14 +233,35 @@
                           u"bold-underline-blink":
                               u"bold-underline-blink-standout",
                           }
-        
+
         # Rebuild focus and non-focus widgets using current properties
-        self._text = (u'%(name)s: %(enabled)s'
-                      % { u"name": self.properties[u"name"],
-                          u"enabled":
-                              (u"enabled"
-                               if self.properties[u"enabled"]
-                               else u"DISABLED")})
+
+        # Base part of a client. Name!
+        base = (u'%(name)s: '
+                      % {u"name": self.properties[u"Name"]})
+        if not self.properties[u"Enabled"]:
+            message = u"DISABLED"
+        elif self.properties[u"ApprovalPending"]:
+            if self.properties[u"ApprovedByDefault"]:
+                message = u"Connection established to client. (d)eny?"
+            else:
+                message = u"Seeks approval to send secret. (a)pprove?"
+        elif self.last_checker_failed:
+            timeout = datetime.timedelta(milliseconds
+                                         = self.properties
+                                         [u"Timeout"])
+            last_ok = isoformat_to_datetime(
+                max((self.properties[u"LastCheckedOK"]
+                     or self.properties[u"Created"]),
+                    self.properties[u"LastEnabled"]))
+            timer = timeout - (datetime.datetime.utcnow() - last_ok)
+            message = (u'A checker has failed! Time until client'
+                       u' gets diabled: %s'
+                           % unicode(timer).rsplit(".", 1)[0])
+        else:
+            message = u"enabled"
+        self._text = "%s%s" % (base, message)
+            
         if not urwid.supports_unicode():
             self._text = self._text.encode("ascii", "replace")
         textlist = [(u"normal", self._text)]
@@ -192,7 +278,15 @@
         if self.update_hook is not None:
             self.update_hook()
     
+    def update_timer(self):
+        "called by gobject"
+        self.update()
+        return True             # Keep calling this
+    
     def delete(self):
+        if self._update_timer_callback_tag is not None:
+            gobject.source_remove(self._update_timer_callback_tag)
+            self._update_timer_callback_tag = None
         if self.delete_hook is not None:
             self.delete_hook(self)
     
@@ -205,19 +299,25 @@
     def keypress(self, (maxcol,), key):
         """Handle keys.
         This overrides the method from urwid.FlowWidget"""
-        if key == u"e" or key == u"+":
-            self.proxy.Enable()
-        elif key == u"d" or key == u"-":
-            self.proxy.Disable()
+        if key == u"+":
+            self.proxy.Enable(dbus_interface = client_interface)
+        elif key == u"-":
+            self.proxy.Disable(dbus_interface = client_interface)
+        elif key == u"a":
+            self.proxy.Approve(dbus.Boolean(True, variant_level=1),
+                               dbus_interface = client_interface)
+        elif key == u"d":
+            self.proxy.Approve(dbus.Boolean(False, variant_level=1),
+                                  dbus_interface = client_interface)
         elif key == u"r" or key == u"_" or key == u"ctrl k":
             self.server_proxy_object.RemoveClient(self.proxy
                                                   .object_path)
         elif key == u"s":
-            self.proxy.StartChecker()
+            self.proxy.StartChecker(dbus_interface = client_interface)
         elif key == u"S":
-            self.proxy.StopChecker()
+            self.proxy.StopChecker(dbus_interface = client_interface)
         elif key == u"C":
-            self.proxy.CheckedOK()
+            self.proxy.CheckedOK(dbus_interface = client_interface)
         # xxx
 #         elif key == u"p" or key == "=":
 #             self.proxy.pause()
@@ -225,6 +325,10 @@
 #             self.proxy.unpause()
 #         elif key == u"RET":
 #             self.open()
+#        elif key == u"+":
+#            self.proxy.Approve(True)
+#        elif key == u"-":
+#            self.proxy.Approve(False)
         else:
             return key
     
@@ -246,7 +350,8 @@
     use them as an excuse to shift focus away from this widget.
     """
     def keypress(self, (maxcol, maxrow), key):
-        ret = super(ConstrainedListBox, self).keypress((maxcol, maxrow), key)
+        ret = super(ConstrainedListBox, self).keypress((maxcol,
+                                                        maxrow), key)
         if ret in (u"up", u"down"):
             return
         return ret
@@ -369,10 +474,13 @@
         Call this when the widget layout needs to change"""
         self.uilist = []
         #self.uilist.append(urwid.ListBox(self.clients))
-        self.uilist.append(urwid.Frame(ConstrainedListBox(self.clients),
+        self.uilist.append(urwid.Frame(ConstrainedListBox(self.
+                                                          clients),
                                        #header=urwid.Divider(),
                                        header=None,
-                                       footer=urwid.Divider(div_char=self.divider)))
+                                       footer=
+                                       urwid.Divider(div_char=
+                                                     self.divider)))
         if self.log_visible:
             self.uilist.append(self.logbox)
             pass
@@ -441,7 +549,7 @@
         if path is None:
             path = client.proxy.object_path
         self.clients_dict[path] = client
-        self.clients.sort(None, lambda c: c.properties[u"name"])
+        self.clients.sort(None, lambda c: c.properties[u"Name"])
         self.refresh()
     
     def remove_client(self, client, path=None):
@@ -522,12 +630,14 @@
                 self.log_message_raw((u"bold",
                                       u"  "
                                       .join((u"Clients:",
-                                             u"e: Enable",
-                                             u"d: Disable",
+                                             u"+: Enable",
+                                             u"-: Disable",
                                              u"r: Remove",
                                              u"s: Start new checker",
                                              u"S: Stop checker",
-                                             u"C: Checker OK"))))
+                                             u"C: Checker OK",
+                                             u"a: Approve",
+                                             u"d: Deny"))))
                 self.refresh()
             elif key == u"tab":
                 if self.topwidget.get_focus() is self.logbox:
@@ -561,6 +671,8 @@
 ui = UserInterface()
 try:
     ui.run()
+except KeyboardInterrupt:
+    ui.screen.stop()
 except Exception, e:
     ui.log_message(unicode(e))
     ui.screen.stop()

=== modified file 'mandos.xml'
--- mandos.xml	2009-12-25 23:13:47 +0000
+++ mandos.xml	2010-09-12 03:00:40 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 "http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos">
-<!ENTITY TIMESTAMP "2009-12-09">
+<!ENTITY TIMESTAMP "2010-09-11">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -353,8 +353,8 @@
     <para>
       The server will by default provide a D-Bus system bus interface.
       This interface will only be accessible by the root user or a
-      Mandos-specific user, if such a user exists.
-      <!-- XXX -->
+      Mandos-specific user, if such a user exists.  For documentation
+      of the D-Bus API, see the file <filename>DBUS-API</filename>.
     </para>
   </refsect1>
   
@@ -458,11 +458,6 @@
       has some security implications, see <xref linkend="clients"/>.
     </para>
     <para>
-      There is currently no way of querying the server of the current
-      status of clients, other than analyzing its <systemitem
-      class="service">syslog</systemitem> output.
-    </para>
-    <para>
       There is no fine-grained control over logging and debug output.
     </para>
     <para>

=== modified file 'plugin-runner.c'
--- plugin-runner.c	2010-03-27 18:39:02 +0000
+++ plugin-runner.c	2010-09-09 22:06:10 +0000
@@ -28,7 +28,7 @@
 #include <stdlib.h>		/* malloc(), exit(), EXIT_SUCCESS,
 				   realloc() */
 #include <stdbool.h>		/* bool, true, false */
-#include <stdio.h>		/* perror, fileno(), fprintf(),
+#include <stdio.h>		/* fileno(), fprintf(),
 				   stderr, STDOUT_FILENO */
 #include <sys/types.h>	        /* DIR, fdopendir(), stat(), struct
 				   stat, waitpid(), WIFEXITED(),
@@ -70,6 +70,8 @@
 #include <inttypes.h>		/* intmax_t, PRIdMAX, strtoimax() */
 #include <sysexits.h>		/* EX_OSERR, EX_USAGE, EX_IOERR,
 				   EX_CONFIG, EX_UNAVAILABLE, EX_OK */
+#include <errno.h> 		/* errno */
+#include <error.h>		/* error() */
 
 #define BUFFER_SIZE 256
 
@@ -266,7 +268,7 @@
 	/* No child processes */
 	break;
       }
-      perror("waitpid");
+      error(0, errno, "waitpid");
     }
     
     /* A child exited, find it in process_list */
@@ -357,13 +359,13 @@
   sigemptyset(&sigchld_action.sa_mask);
   ret = sigaddset(&sigchld_action.sa_mask, SIGCHLD);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     exitstatus = EX_OSERR;
     goto fallback;
   }
   ret = sigaction(SIGCHLD, &sigchld_action, &old_sigchld_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     exitstatus = EX_OSERR;
     goto fallback;
   }
@@ -599,7 +601,7 @@
   case ENOMEM:
   default:
     errno = ret;
-    perror("argp_parse");
+    error(0, errno, "argp_parse");
     exitstatus = EX_OSERR;
     goto fallback;
   case EINVAL:
@@ -626,7 +628,7 @@
     custom_argc = 1;
     custom_argv = malloc(sizeof(char*) * 2);
     if(custom_argv == NULL){
-      perror("malloc");
+      error(0, errno, "malloc");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -649,7 +651,7 @@
 	}
 	new_arg = strdup(p);
 	if(new_arg == NULL){
-	  perror("strdup");
+	  error(0, errno, "strdup");
 	  exitstatus = EX_OSERR;
 	  free(org_line);
 	  goto fallback;
@@ -659,7 +661,7 @@
 	custom_argv = realloc(custom_argv, sizeof(char *)
 			      * ((unsigned int) custom_argc + 1));
 	if(custom_argv == NULL){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  exitstatus = EX_OSERR;
 	  free(org_line);
 	  goto fallback;
@@ -672,7 +674,7 @@
       ret = fclose(conffp);
     } while(ret == EOF and errno == EINTR);
     if(ret == EOF){
-      perror("fclose");
+      error(0, errno, "fclose");
       exitstatus = EX_IOERR;
       goto fallback;
     }
@@ -681,7 +683,7 @@
     /* Check for harmful errors and go to fallback. Other errors might
        not affect opening plugins */
     if(errno == EMFILE or errno == ENFILE or errno == ENOMEM){
-      perror("fopen");
+      error(0, errno, "fopen");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -698,7 +700,7 @@
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      error(0, errno, "argp_parse");
       exitstatus = EX_OSERR;
       goto fallback;
     case EINVAL:
@@ -718,7 +720,7 @@
   case ENOMEM:
   default:
     errno = ret;
-    perror("argp_parse");
+    error(0, errno, "argp_parse");
     exitstatus = EX_OSERR;
     goto fallback;
   case EINVAL:
@@ -743,11 +745,11 @@
   /* Strip permissions down to nobody */
   setgid(gid);
   if(ret == -1){
-    perror("setgid");
+    error(0, errno, "setgid");
   }
   ret = setuid(uid);
   if(ret == -1){
-    perror("setuid");
+    error(0, errno, "setuid");
   }
   
   /* Open plugin directory with close_on_exec flag */
@@ -771,7 +773,7 @@
 		    );
     }
     if(dir_fd == -1){
-      perror("Could not open plugin dir");
+      error(0, errno, "Could not open plugin dir");
       exitstatus = EX_UNAVAILABLE;
       goto fallback;
     }
@@ -780,7 +782,7 @@
   /* Set the FD_CLOEXEC flag on the directory */
     ret = set_cloexec_flag(dir_fd);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
@@ -789,7 +791,7 @@
     
     dir = fdopendir(dir_fd);
     if(dir == NULL){
-      perror("Could not open plugin dir");
+      error(0, errno, "Could not open plugin dir");
       TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
@@ -807,7 +809,7 @@
     /* All directory entries have been processed */
     if(dirst == NULL){
       if(errno == EBADF){
-	perror("readdir");
+	error(0, errno, "readdir");
 	exitstatus = EX_IOERR;
 	goto fallback;
       }
@@ -870,13 +872,13 @@
 					     dirst->d_name));
     }
     if(ret < 0){
-      perror("asprintf");
+      error(0, errno, "asprintf");
       continue;
     }
     
     ret = (int)TEMP_FAILURE_RETRY(stat(filename, &st));
     if(ret == -1){
-      perror("stat");
+      error(0, errno, "stat");
       free(filename);
       continue;
     }
@@ -894,7 +896,7 @@
     
     plugin *p = getplugin(dirst->d_name);
     if(p == NULL){
-      perror("getplugin");
+      error(0, errno, "getplugin");
       free(filename);
       continue;
     }
@@ -912,13 +914,13 @@
       if(g != NULL){
 	for(char **a = g->argv + 1; *a != NULL; a++){
 	  if(not add_argument(p, *a)){
-	    perror("add_argument");
+	    error(0, errno, "add_argument");
 	  }
 	}
 	/* Add global environment variables */
 	for(char **e = g->environ; *e != NULL; e++){
 	  if(not add_environment(p, *e, false)){
-	    perror("add_environment");
+	    error(0, errno, "add_environment");
 	  }
 	}
       }
@@ -929,7 +931,7 @@
     if(p->environ[0] != NULL){
       for(char **e = environ; *e != NULL; e++){
 	if(not add_environment(p, *e, false)){
-	  perror("add_environment");
+	  error(0, errno, "add_environment");
 	}
       }
     }
@@ -937,20 +939,20 @@
     int pipefd[2];
     ret = (int)TEMP_FAILURE_RETRY(pipe(pipefd));
     if(ret == -1){
-      perror("pipe");
+      error(0, errno, "pipe");
       exitstatus = EX_OSERR;
       goto fallback;
     }
     /* Ask OS to automatic close the pipe on exec */
     ret = set_cloexec_flag(pipefd[0]);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       exitstatus = EX_OSERR;
       goto fallback;
     }
     ret = set_cloexec_flag(pipefd[1]);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -959,7 +961,7 @@
 					      &sigchld_action.sa_mask,
 					      NULL));
     if(ret < 0){
-      perror("sigprocmask");
+      error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -969,7 +971,7 @@
       pid = fork();
     } while(pid == -1 and errno == EINTR);
     if(pid == -1){
-      perror("fork");
+      error(0, errno, "fork");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -977,18 +979,18 @@
       /* this is the child process */
       ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
       if(ret < 0){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	_exit(EX_OSERR);
       }
       ret = sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
       if(ret < 0){
-	perror("sigprocmask");
+	error(0, errno, "sigprocmask");
 	_exit(EX_OSERR);
       }
       
       ret = dup2(pipefd[1], STDOUT_FILENO); /* replace our stdout */
       if(ret == -1){
-	perror("dup2");
+	error(0, errno, "dup2");
 	_exit(EX_OSERR);
       }
       
@@ -999,12 +1001,12 @@
       }
       if(p->environ[0] == NULL){
 	if(execv(filename, p->argv) < 0){
-	  perror("execv");
+	  error(0, errno, "execv for %s", filename);
 	  _exit(EX_OSERR);
 	}
       } else {
 	if(execve(filename, p->argv, p->environ) < 0){
-	  perror("execve");
+	  error(0, errno, "execve for %s", filename);
 	  _exit(EX_OSERR);
 	}
       }
@@ -1016,12 +1018,12 @@
     free(filename);
     plugin *new_plugin = getplugin(dirst->d_name);
     if(new_plugin == NULL){
-      perror("getplugin");
+      error(0, errno, "getplugin");
       ret = (int)(TEMP_FAILURE_RETRY
 		  (sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask,
 			       NULL)));
       if(ret < 0){
-        perror("sigprocmask");
+        error(0, errno, "sigprocmask");
       }
       exitstatus = EX_OSERR;
       goto fallback;
@@ -1036,7 +1038,7 @@
 					      &sigchld_action.sa_mask,
 					      NULL));
     if(ret < 0){
-      perror("sigprocmask");
+      error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1069,7 +1071,7 @@
     fd_set rfds = rfds_all;
     int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
     if(select_ret == -1 and errno != EINTR){
-      perror("select");
+      error(0, errno, "select");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1111,7 +1113,7 @@
 					 &sigchld_action.sa_mask,
 					 NULL));
 	  if(ret < 0){
-	    perror("sigprocmask");
+	    error(0, errno, "sigprocmask");
 	    exitstatus = EX_OSERR;
 	    goto fallback;
 	  }
@@ -1125,7 +1127,7 @@
 		      (sigprocmask(SIG_UNBLOCK,
 				   &sigchld_action.sa_mask, NULL)));
 	  if(ret < 0){
-	    perror("sigprocmask");
+	    error(0, errno, "sigprocmask");
 	    exitstatus = EX_OSERR;
 	    goto fallback;
 	  }
@@ -1142,7 +1144,7 @@
 	bool bret = print_out_password(proc->buffer,
 				       proc->buffer_length);
 	if(not bret){
-	  perror("print_out_password");
+	  error(0, errno, "print_out_password");
 	  exitstatus = EX_IOERR;
 	}
 	goto fallback;
@@ -1161,7 +1163,7 @@
 	proc->buffer = realloc(proc->buffer, proc->buffer_size
 			       + (size_t) BUFFER_SIZE);
 	if(proc->buffer == NULL){
-	  perror("malloc");
+	  error(0, errno, "malloc");
 	  exitstatus = EX_OSERR;
 	  goto fallback;
 	}
@@ -1204,7 +1206,7 @@
     }
     bret = print_out_password(passwordbuffer, len);
     if(not bret){
-      perror("print_out_password");
+      error(0, errno, "print_out_password");
       exitstatus = EX_IOERR;
     }
   }
@@ -1212,7 +1214,7 @@
   /* Restore old signal handler */
   ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     exitstatus = EX_OSERR;
   }
   
@@ -1234,7 +1236,7 @@
       ret = kill(p->pid, SIGTERM);
       if(ret == -1 and errno != ESRCH){
 	/* Set-uid proccesses might not get closed */
-	perror("kill");
+	error(0, errno, "kill");
       }
     }
   }
@@ -1244,7 +1246,7 @@
     ret = wait(NULL);
   } while(ret >= 0);
   if(errno != ECHILD){
-    perror("wait");
+    error(0, errno, "wait");
   }
   
   free_plugin_list();

=== modified file 'plugins.d/askpass-fifo.c'
--- plugins.d/askpass-fifo.c	2009-10-18 16:09:05 +0000
+++ plugins.d/askpass-fifo.c	2010-09-07 16:48:58 +0000
@@ -31,7 +31,7 @@
 				   ENOENT, EEXIST, EFAULT, EMFILE,
 				   ENFILE, ENOMEM, EBADF, EINVAL, EIO,
 				   EISDIR, EFBIG */
-#include <stdio.h>		/* perror() */
+#include <error.h>		/* error() */
 #include <stdlib.h>		/* EXIT_FAILURE, NULL, size_t, free(),
 				   realloc(), EXIT_SUCCESS */
 #include <fcntl.h>		/* open(), O_RDONLY */
@@ -51,7 +51,7 @@
   ret = mkfifo(passfifo, S_IRUSR | S_IWUSR);
   if(ret == -1){
     int e = errno;
-    perror("mkfifo");
+    error(0, errno, "mkfifo");
     switch(e){
     case EACCES:
     case ENOTDIR:
@@ -73,7 +73,7 @@
   int fifo_fd = open(passfifo, O_RDONLY);
   if(fifo_fd == -1){
     int e = errno;
-    perror("open");
+    error(0, errno, "open");
     switch(e){
     case EACCES:
     case ENOENT:
@@ -101,7 +101,7 @@
       if(buf_len + blocksize > buf_allocated){
 	char *tmp = realloc(buf, buf_allocated + blocksize);
 	if(tmp == NULL){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  free(buf);
 	  return EX_OSERR;
 	}
@@ -113,7 +113,7 @@
 	int e = errno;
 	free(buf);
 	errno = e;
-	perror("read");
+	error(0, errno, "read");
 	switch(e){
 	case EBADF:
 	case EFAULT:
@@ -141,7 +141,7 @@
       int e = errno;
       free(buf);
       errno = e;
-      perror("write");
+      error(0, errno, "write");
       switch(e){
       case EBADF:
       case EFAULT:
@@ -161,7 +161,7 @@
   ret = close(STDOUT_FILENO);
   if(ret == -1){
     int e = errno;
-    perror("close");
+    error(0, errno, "close");
     switch(e){
     case EBADF:
       return EX_OSFILE;

=== modified file 'plugins.d/mandos-client.c'
--- plugins.d/mandos-client.c	2009-11-01 00:10:28 +0000
+++ plugins.d/mandos-client.c	2010-09-09 18:16:14 +0000
@@ -127,6 +127,8 @@
 static const char mandos_protocol_version[] = "1";
 const char *argp_program_version = "mandos-client " VERSION;
 const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
+static const char sys_class_net[] = "/sys/class/net";
+char *connect_to = NULL;
 
 /* Used for passing in values through the Avahi callback functions */
 typedef struct {
@@ -1020,6 +1022,101 @@
   errno = old_errno;
 }
 
+/* 
+ * This function determines if a directory entry in /sys/class/net
+ * corresponds to an acceptable network device.
+ * (This function is passed to scandir(3) as a filter function.)
+ */
+int good_interface(const struct dirent *if_entry){
+  ssize_t ssret;
+  char *flagname = NULL;
+  int ret = asprintf(&flagname, "%s/%s/flags", sys_class_net,
+		     if_entry->d_name);
+  if(ret < 0){
+    perror("asprintf");
+    return 0;
+  }
+  if(if_entry->d_name[0] == '.'){
+    return 0;
+  }
+  int flags_fd = (int)TEMP_FAILURE_RETRY(open(flagname, O_RDONLY));
+  if(flags_fd == -1){
+    perror("open");
+    return 0;
+  }
+  typedef short ifreq_flags;	/* ifreq.ifr_flags in netdevice(7) */
+  /* read line from flags_fd */
+  ssize_t to_read = (sizeof(ifreq_flags)*2)+3; /* "0x1003\n" */
+  char *flagstring = malloc((size_t)to_read+1); /* +1 for final \0 */
+  flagstring[(size_t)to_read] = '\0';
+  if(flagstring == NULL){
+    perror("malloc");
+    close(flags_fd);
+    return 0;
+  }
+  while(to_read > 0){
+    ssret = (ssize_t)TEMP_FAILURE_RETRY(read(flags_fd, flagstring,
+					     (size_t)to_read));
+    if(ssret == -1){
+      perror("read");
+      free(flagstring);
+      close(flags_fd);
+      return 0;
+    }
+    to_read -= ssret;
+    if(ssret == 0){
+      break;
+    }
+  }
+  close(flags_fd);
+  intmax_t tmpmax;
+  char *tmp;
+  errno = 0;
+  tmpmax = strtoimax(flagstring, &tmp, 0);
+  if(errno != 0 or tmp == flagstring or (*tmp != '\0'
+					 and not (isspace(*tmp)))
+     or tmpmax != (ifreq_flags)tmpmax){
+    if(debug){
+      fprintf(stderr, "Invalid flags \"%s\" for interface \"%s\"\n",
+	      flagstring, if_entry->d_name);
+    }
+    free(flagstring);
+    return 0;
+  }
+  free(flagstring);
+  ifreq_flags flags = (ifreq_flags)tmpmax;
+  /* Reject the loopback device */
+  if(flags & IFF_LOOPBACK){
+    if(debug){
+      fprintf(stderr, "Rejecting loopback interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  /* Accept point-to-point devices only if connect_to is specified */
+  if(connect_to != NULL and (flags & IFF_POINTOPOINT)){
+    if(debug){
+      fprintf(stderr, "Accepting point-to-point interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 1;
+  }
+  /* Otherwise, reject non-broadcast-capable devices */
+  if(not (flags & IFF_BROADCAST)){
+    if(debug){
+      fprintf(stderr, "Rejecting non-broadcast interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  /* Accept this device */
+  if(debug){
+    fprintf(stderr, "Interface \"%s\" is acceptable\n",
+	    if_entry->d_name);
+  }
+  return 1;
+}
+
 int main(int argc, char *argv[]){
   AvahiSServiceBrowser *sb = NULL;
   int error;
@@ -1027,13 +1124,12 @@
   intmax_t tmpmax;
   char *tmp;
   int exitcode = EXIT_SUCCESS;
-  const char *interface = "eth0";
+  const char *interface = "";
   struct ifreq network;
   int sd = -1;
   bool take_down_interface = false;
   uid_t uid;
   gid_t gid;
-  char *connect_to = NULL;
   char tempdir[] = "/tmp/mandosXXXXXX";
   bool tempdir_created = false;
   AvahiIfIndex if_index = AVAHI_IF_UNSPEC;
@@ -1196,6 +1292,31 @@
   if(not debug){
     avahi_set_log_function(empty_log);
   }
+
+  if(interface[0] == '\0'){
+    struct dirent **direntries;
+    ret = scandir(sys_class_net, &direntries, good_interface,
+		  alphasort);
+    if(ret >= 1){
+      /* Pick the first good interface */
+      interface = strdup(direntries[0]->d_name);
+      if(debug){
+	fprintf(stderr, "Using interface \"%s\"\n", interface);
+      }
+      if(interface == NULL){
+	perror("malloc");
+	free(direntries);
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      free(direntries);
+    } else {
+      free(direntries);
+      fprintf(stderr, "Could not find a network interface\n");
+      exitcode = EXIT_FAILURE;
+      goto end;
+    }
+  }
   
   /* Initialize Avahi early so avahi_simple_poll_quit() can be called
      from the signal handler */
@@ -1272,7 +1393,7 @@
   }
   
   /* If the interface is down, bring it up */
-  if(interface[0] != '\0'){
+  if(strcmp(interface, "none") != 0){
     if_index = (AvahiIfIndex) if_nametoindex(interface);
     if(if_index == 0){
       fprintf(stderr, "No such interface: \"%s\"\n", interface);
@@ -1349,7 +1470,7 @@
       ret = ioctl(sd, SIOCSIFFLAGS, &network);
       if(ret == -1){
 	take_down_interface = false;
-	perror("ioctl SIOCSIFFLAGS");
+	perror("ioctl SIOCSIFFLAGS +IFF_UP");
 	exitcode = EX_OSERR;
 #ifdef __linux__
 	if(restore_loglevel){
@@ -1624,7 +1745,7 @@
 	network.ifr_flags &= ~(short)IFF_UP; /* clear flag */
 	ret = ioctl(sd, SIOCSIFFLAGS, &network);
 	if(ret == -1){
-	  perror("ioctl SIOCSIFFLAGS");
+	  perror("ioctl SIOCSIFFLAGS -IFF_UP");
 	}
       }
       ret = (int)TEMP_FAILURE_RETRY(close(sd));

=== modified file 'plugins.d/mandos-client.xml'
--- plugins.d/mandos-client.xml	2009-02-09 02:01:13 +0000
+++ plugins.d/mandos-client.xml	2010-08-10 19:08:24 +0000
@@ -202,8 +202,9 @@
 	<listitem>
 	  <para>
 	    Network interface that will be brought up and scanned for
-	    Mandos servers to connect to.  The default is
-	    <quote><literal>eth0</literal></quote>.
+	    Mandos servers to connect to.  The default is the empty
+	    string, which will automatically choose an appropriate
+	    interface.
 	  </para>
 	  <para>
 	    If the <option>--connect</option> option is used, this
@@ -220,10 +221,11 @@
 	    by this program.
 	  </para>
 	  <para>
-	    <replaceable>NAME</replaceable> can be the empty string;
-	    this will not use any specific interface, and will not
-	    bring up an interface on startup.  This is not
-	    recommended, and only meant for advanced users.
+	    <replaceable>NAME</replaceable> can be the string
+	    <quote><literal>none</literal></quote>; this will not use
+	    any specific interface, and will not bring up an interface
+	    on startup.  This is not recommended, and only meant for
+	    advanced users.
 	  </para>
 	</listitem>
       </varlistentry>

=== modified file 'plugins.d/password-prompt.c'
--- plugins.d/password-prompt.c	2010-03-27 18:39:02 +0000
+++ plugins.d/password-prompt.c	2010-09-07 16:48:58 +0000
@@ -39,11 +39,12 @@
 #include <stdlib.h>		/* EXIT_SUCCESS, EXIT_FAILURE,
 				   getenv() */
 #include <stdio.h>		/* fprintf(), stderr, getline(),
-				   stdin, feof(), perror(), fputc()
+				   stdin, feof(), fputc()
 				*/
 #include <errno.h>		/* errno, EBADF, ENOTTY, EINVAL,
 				   EFAULT, EFBIG, EIO, ENOSPC, EINTR
 				*/
+#include <error.h>		/* error() */
 #include <iso646.h>		/* or, not */
 #include <stdbool.h>		/* bool, false, true */
 #include <string.h> 		/* strlen, rindex */
@@ -70,7 +71,8 @@
 }
 
 int main(int argc, char **argv){
-  ssize_t ret;
+  ssize_t sret;
+  int ret;
   size_t n;
   struct termios t_new, t_old;
   char *buffer = NULL;
@@ -139,7 +141,7 @@
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      error(0, errno, "argp_parse");
       return EX_OSERR;
     case EINVAL:
       return EX_USAGE;
@@ -155,7 +157,7 @@
   
   if(tcgetattr(STDIN_FILENO, &t_old) != 0){
     int e = errno;
-    perror("tcgetattr");
+    error(0, errno, "tcgetattr");
     switch(e){
     case EBADF:
     case ENOTTY:
@@ -168,17 +170,17 @@
   sigemptyset(&new_action.sa_mask);
   ret = sigaddset(&new_action.sa_mask, SIGINT);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   ret = sigaddset(&new_action.sa_mask, SIGHUP);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   ret = sigaddset(&new_action.sa_mask, SIGTERM);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   /* Need to check if the handler is SIG_IGN before handling:
@@ -187,37 +189,37 @@
   */
   ret = sigaction(SIGINT, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGINT, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
   ret = sigaction(SIGHUP, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGHUP, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
   ret = sigaction(SIGTERM, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGTERM, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
@@ -231,7 +233,7 @@
   t_new.c_lflag &= ~(tcflag_t)ECHO;
   if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_new) != 0){
     int e = errno;
-    perror("tcsetattr-echo");
+    error(0, errno, "tcsetattr-echo");
     switch(e){
     case EBADF:
     case ENOTTY:
@@ -286,11 +288,11 @@
 	}
       }
     }
-    ret = getline(&buffer, &n, stdin);
-    if(ret > 0){
+    sret = getline(&buffer, &n, stdin);
+    if(sret > 0){
       status = EXIT_SUCCESS;
       /* Make n = data size instead of allocated buffer size */
-      n = (size_t)ret;
+      n = (size_t)sret;
       /* Strip final newline */
       if(n > 0 and buffer[n-1] == '\n'){
 	buffer[n-1] = '\0';	/* not strictly necessary */
@@ -298,10 +300,10 @@
       }
       size_t written = 0;
       while(written < n){
-	ret = write(STDOUT_FILENO, buffer + written, n - written);
-	if(ret < 0){
+	sret = write(STDOUT_FILENO, buffer + written, n - written);
+	if(sret < 0){
 	  int e = errno;
-	  perror("write");
+	  error(0, errno, "write");
 	  switch(e){
 	  case EBADF:
 	  case EFAULT:
@@ -318,12 +320,12 @@
 	  }
 	  break;
 	}
-	written += (size_t)ret;
+	written += (size_t)sret;
       }
-      ret = close(STDOUT_FILENO);
-      if(ret == -1){
+      sret = close(STDOUT_FILENO);
+      if(sret == -1){
 	int e = errno;
-	perror("close");
+	error(0, errno, "close");
 	switch(e){
 	case EBADF:
 	  status = EX_OSFILE;
@@ -336,10 +338,10 @@
       }
       break;
     }
-    if(ret < 0){
+    if(sret < 0){
       int e = errno;
       if(errno != EINTR and not feof(stdin)){
-	perror("getline");
+	error(0, errno, "getline");
 	switch(e){
 	case EBADF:
 	  status = EX_UNAVAILABLE;
@@ -352,7 +354,7 @@
 	break;
       }
     }
-    /* if(ret == 0), then the only sensible thing to do is to retry to
+    /* if(sret == 0), then the only sensible thing to do is to retry to
        read from stdin */
     fputc('\n', stderr);
     if(debug and not quit_now){
@@ -368,7 +370,7 @@
     fprintf(stderr, "Restoring terminal attributes\n");
   }
   if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_old) != 0){
-    perror("tcsetattr+echo");
+    error(0, errno, "tcsetattr+echo");
   }
   
   if(quit_now){
@@ -376,7 +378,7 @@
     old_action.sa_handler = SIG_DFL;
     ret = sigaction(signal_received, &old_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     raise(signal_received);
   }

=== added file 'plugins.d/plymouth.c'
--- plugins.d/plymouth.c	1970-01-01 00:00:00 +0000
+++ plugins.d/plymouth.c	2010-09-07 18:48:56 +0000
@@ -0,0 +1,405 @@
+#define _GNU_SOURCE		/* asprintf(), TEMP_FAILURE_RETRY() */
+#include <signal.h>		/* sig_atomic_t, struct sigaction,
+				   sigemptyset(), sigaddset(), SIGINT,
+				   SIGHUP, SIGTERM, sigaction(),
+				   kill(), SIG_IGN */
+#include <stdbool.h>		/* bool, false, true */
+#include <fcntl.h>		/* open(), O_RDONLY */
+#include <iso646.h>		/* and, or, not*/
+#include <sys/types.h>		/* size_t, ssize_t, pid_t, struct dirent,
+				   waitpid() */
+#include <sys/wait.h>		/* waitpid() */
+#include <stddef.h>		/* NULL */
+#include <string.h>		/* strchr(), memcmp() */
+#include <stdio.h>		/* asprintf(), perror(), fopen(), fscanf() */
+#include <unistd.h>		/* close(), readlink(), read(), fork()
+				   setsid(), chdir(), dup2()
+				   STDERR_FILENO, execv(), access() */
+#include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
+#include <dirent.h>		/* scandir(), alphasort() */
+#include <inttypes.h>		/* intmax_t, strtoumax(), SCNuMAX */
+#include <sys/stat.h>		/* struct stat, lstat() */
+#include <sysexits.h>		/* EX_OSERR */
+#include <error.h>		/* error() */
+#include <errno.h>		/* TEMP_FAILURE_RETRY */
+#include <stdarg.h>
+
+sig_atomic_t interrupted_by_signal = 0;
+const char plymouth_pid[] = "/dev/.initramfs/plymouth.pid";
+const char plymouth_path[] = "/bin/plymouth";
+const char plymouthd_path[] = "/sbin/plymouthd";
+const char *plymouthd_default_argv[] = {"/sbin/plymouthd", "--mode=boot",
+					"--attach-to-session",
+					"--pid-file=/dev/.initramfs/plymouth.pid",
+					NULL };
+
+static void termination_handler(__attribute__((unused))int signum){
+  if(interrupted_by_signal){
+    return;
+  }
+  interrupted_by_signal = 1;
+}
+
+/* Create prompt string */
+char *makeprompt(void){
+  int ret = 0;
+  char *prompt;
+  const char *const cryptsource = getenv("cryptsource");
+  const char *const crypttarget = getenv("crypttarget");
+  const char prompt_start[] = "Enter passphrase to unlock the disk";
+  
+  if(cryptsource == NULL){
+    if(crypttarget == NULL){
+      ret = asprintf(&prompt, "%s: ", prompt_start);
+    } else {
+      ret = asprintf(&prompt, "%s (%s): ", prompt_start,
+		     crypttarget);
+    }
+  } else {
+    if(crypttarget == NULL){
+      ret = asprintf(&prompt, "%s %s: ", prompt_start, cryptsource);
+    } else {
+      ret = asprintf(&prompt, "%s %s (%s): ", prompt_start,
+		     cryptsource, crypttarget);
+    }
+  }
+  if(ret == -1){
+    return NULL;
+  }
+  return prompt;
+}
+
+void kill_and_wait(pid_t pid){
+  TEMP_FAILURE_RETRY(kill(pid, SIGTERM));
+  TEMP_FAILURE_RETRY(waitpid(pid, NULL, 0));
+}
+
+bool become_a_daemon(void){
+  int ret = setuid(geteuid());
+  if(ret == -1){
+    error(0, errno, "setuid");
+  }
+    
+  setsid();
+  ret = chdir("/");
+  if(ret == -1){
+    error(0, errno, "chdir");
+    return false;
+  }
+  ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
+  if(ret == -1){
+    error(0, errno, "dup2");
+    return false;
+  }
+  return true;
+}
+
+bool exec_and_wait(pid_t *pid_return, const char *path,
+		   const char **argv, bool interruptable,
+		   bool daemonize){
+  int status;
+  int ret;
+  pid_t pid;
+  pid = fork();
+  if(pid == -1){
+    error(0, errno, "fork");
+    return false;
+  }
+  if(pid == 0){
+    /* Child */
+    if(daemonize){
+      if(not become_a_daemon()){
+	_exit(EX_OSERR);
+      }
+    }
+
+    char **new_argv = NULL;
+    char *tmp;
+    int i = 0;
+    for (; argv[i]!=(char *)NULL; i++){
+      tmp = realloc(new_argv, sizeof(const char *) * ((size_t)i + 1));
+      if (tmp == NULL){
+	error(0, errno, "realloc");
+	free(new_argv);
+	_exit(EXIT_FAILURE);
+      }
+      new_argv = (char **)tmp;
+      new_argv[i] = strdup(argv[i]);
+    }
+    new_argv[i] = (char *) NULL;
+    
+    execv(path, (char *const *)new_argv);
+    error(0, errno, "execv");
+    _exit(EXIT_FAILURE);
+  }
+  if(pid_return != NULL){
+    *pid_return = pid;
+  }
+  do {
+    ret = waitpid(pid, &status, 0);
+  } while(ret == -1 and errno == EINTR
+	  and ((not interrupted_by_signal)
+	       or (not interruptable)));
+  if(interrupted_by_signal and interruptable){
+    return false;
+  }
+  if(ret == -1){
+    error(0, errno, "waitpid");
+    return false;
+  }
+  if(WIFEXITED(status) and WEXITSTATUS(status) == 0){
+    return true;
+  }
+  return false;
+}
+
+int is_plymouth(const struct dirent *proc_entry){
+  int ret;
+  {
+    uintmax_t maxvalue;
+    char *tmp;
+    errno = 0;
+    maxvalue = strtoumax(proc_entry->d_name, &tmp, 10);
+
+    if(errno != 0 or *tmp != '\0' or maxvalue != (uintmax_t)((pid_t)maxvalue)){
+      return 0;
+    }
+  }
+  char exe_target[sizeof(plymouth_path)];
+  char *exe_link;
+  ret = asprintf(&exe_link, "/proc/%s/exe", proc_entry->d_name);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    return 0;
+  }
+
+  struct stat exe_stat;
+  ret = lstat(exe_link, &exe_stat);
+  if(ret == -1){
+    free(exe_link);
+    if(errno != ENOENT){
+      error(0, errno, "lstat");
+    }
+    return 0;
+  }
+  
+  if(not S_ISLNK(exe_stat.st_mode)
+     or exe_stat.st_uid != 0
+     or exe_stat.st_gid != 0){
+    free(exe_link);
+    return 0;
+  }
+
+  ssize_t sret = readlink(exe_link, exe_target, sizeof(exe_target));
+  free(exe_link);
+  if((sret != (ssize_t)sizeof(plymouth_path)-1) or
+      (memcmp(plymouth_path, exe_target,
+	      sizeof(plymouth_path)-1) != 0)){
+    return 0;
+  }
+  return 1;
+}
+
+pid_t get_pid(void){
+  int ret;
+  FILE *pidfile = fopen(plymouth_pid, "r");
+  uintmax_t maxvalue = 0;
+  if(pidfile != NULL){
+    ret = fscanf(pidfile, "%" SCNuMAX, &maxvalue);
+    if(ret != 1){
+      maxvalue = 0;
+    }
+    fclose(pidfile);
+  }
+  if(maxvalue == 0){
+    struct dirent **direntries;
+    ret = scandir("/proc", &direntries, is_plymouth, alphasort);
+    sscanf(direntries[0]->d_name, "%" SCNuMAX, &maxvalue);
+  }
+  pid_t pid;
+  pid = (pid_t)maxvalue;
+  if((uintmax_t)pid == maxvalue){
+    return pid;
+  }
+  
+  return 0;
+}
+
+const char **getargv(pid_t pid){
+  int cl_fd;
+  char *cmdline_filename;
+  ssize_t sret;
+  int ret;
+  
+  ret = asprintf(&cmdline_filename, "/proc/%" PRIuMAX "/cmdline",
+		 (uintmax_t)pid);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    return NULL;
+  }
+  
+  /* Open /proc/<pid>/cmdline  */
+  cl_fd = open(cmdline_filename, O_RDONLY);
+  free(cmdline_filename);
+  if(cl_fd == -1){
+    error(0, errno, "open");
+    return NULL;
+  }
+  
+  size_t cmdline_allocated = 0;
+  size_t cmdline_len = 0;
+  char *cmdline = NULL;
+  char *tmp;
+  const size_t blocksize = 1024;
+  do {
+    /* Allocate more space? */
+    if(cmdline_len + blocksize > cmdline_allocated){
+      tmp = realloc(cmdline, cmdline_allocated + blocksize);
+      if(tmp == NULL){
+	error(0, errno, "realloc");
+	free(cmdline);
+	close(cl_fd);
+	return NULL;
+      }
+      cmdline = tmp;
+      cmdline_allocated += blocksize;
+    }
+    
+    /* Read data */
+    sret = read(cl_fd, cmdline + cmdline_len,
+		cmdline_allocated - cmdline_len);
+    if(sret == -1){
+      error(0, errno, "read");
+      free(cmdline);
+      close(cl_fd);
+      return NULL;
+    }
+    cmdline_len += (size_t)sret;
+  } while(sret != 0);
+  ret = close(cl_fd);
+  if(ret == -1){
+    error(0, errno, "close");
+    free(cmdline);
+    return NULL;
+  }
+  
+  /* we got cmdline and cmdline_len, ignore rest... */
+  const char **argv = NULL;
+  size_t argv_size = 0;
+  for(char *arg = cmdline; arg-cmdline < (ssize_t)cmdline_len;
+      arg = strchr(arg, '\0')+1){
+    tmp = realloc(argv, ((++argv_size)+1)*sizeof(char *));
+    if(tmp == NULL){
+      error(0, errno, "realloc");
+      free(argv);
+      return NULL;
+    }
+    argv = (const char **)tmp;
+    argv[argv_size-1] = arg;
+  }
+  argv[argv_size] = NULL;
+  return argv;
+}
+
+int main(__attribute__((unused))int argc,
+	 __attribute__((unused))char **argv){
+  char *prompt;
+  char *prompt_arg;
+  pid_t plymouth_command_pid;
+  int ret;
+  bool bret;
+
+  /* test -x /bin/plymouth */
+  ret = access(plymouth_path, X_OK);
+  if(ret == -1){
+    exit(EXIT_FAILURE);
+  }
+
+  { /* Add signal handlers */
+    struct sigaction old_action,
+      new_action = { .sa_handler = termination_handler,
+		     .sa_flags = 0 };
+    sigemptyset(&new_action.sa_mask);
+    for(int *sig = (int[]){ SIGINT, SIGHUP, SIGTERM, 0 }; *sig != 0; sig++){
+      ret = sigaddset(&new_action.sa_mask, *sig);
+      if(ret == -1){
+	error(0, errno, "sigaddset");
+	exit(EX_OSERR);
+      }
+      ret = sigaction(*sig, NULL, &old_action);
+      if(ret == -1){
+	error(0, errno, "sigaction");
+	exit(EX_OSERR);
+      }
+      if(old_action.sa_handler != SIG_IGN){
+	ret = sigaction(*sig, &new_action, NULL);
+	if(ret == -1){
+	  error(0, errno, "sigaction");
+	  exit(EX_OSERR);
+	}
+      }
+    }
+  }
+    
+  /* plymouth --ping */
+  bret = exec_and_wait(&plymouth_command_pid, plymouth_path,
+		       (const char *[]){ (const char *)plymouth_path, (const char *)"--ping", (const char *)NULL},
+		       true, false);
+  if(not bret){
+    if(interrupted_by_signal){
+      kill_and_wait(plymouth_command_pid);
+    }
+    exit(EXIT_FAILURE);
+  }
+  
+  prompt = makeprompt();
+  ret = asprintf(&prompt_arg, "--prompt=%s", prompt);
+  free(prompt);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    exit(EXIT_FAILURE);
+  }
+  
+  /* plymouth ask-for-password --prompt="$prompt" */
+  bret = exec_and_wait(&plymouth_command_pid, plymouth_path,
+		       (const char *[]){plymouth_path, "ask-for-password", prompt_arg, NULL},
+		       true, false);
+  free(prompt_arg);
+  if(not bret){
+    if(interrupted_by_signal){
+      kill_and_wait(plymouth_command_pid);
+    } else {
+      exit(EXIT_FAILURE);
+    }
+  }
+  
+  if(bret){
+    exit(EXIT_SUCCESS);
+  }
+  
+  const char **plymouthd_argv = NULL;
+  pid_t pid = get_pid();
+  if(pid == 0){
+    error(0, 0, "plymouthd pid not found");
+  } else {
+    plymouthd_argv = getargv(pid);
+  }
+  if(plymouthd_argv == NULL){
+    plymouthd_argv = plymouthd_default_argv;
+  }
+  
+  bret = exec_and_wait(NULL, plymouth_path,
+  		       (const char *[]){plymouth_path, "quit", NULL}, false, false);
+  if(not bret){
+    exit(EXIT_FAILURE);
+  }
+  bret = exec_and_wait(NULL, plymouthd_path, plymouthd_argv, false, true);
+  if(not bret){
+    exit(EXIT_FAILURE);
+  }
+  exec_and_wait(NULL, plymouth_path,
+  		(const char *[]){ plymouth_path, "show-splash", NULL }, false, false);
+  exit(EXIT_FAILURE);
+}

=== modified file 'plugins.d/splashy.c'
--- plugins.d/splashy.c	2009-10-18 16:09:05 +0000
+++ plugins.d/splashy.c	2010-09-07 16:48:58 +0000
@@ -29,7 +29,7 @@
 				   SIG_IGN, kill(), SIGKILL */
 #include <stddef.h>		/* NULL */
 #include <stdlib.h>		/* getenv() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf() */
 #include <stdlib.h>		/* EXIT_FAILURE, free(),
 				   EXIT_SUCCESS */
 #include <sys/types.h>		/* pid_t, DIR, struct dirent,
@@ -49,6 +49,7 @@
 				   E2BIG, EFAULT, EIO, ETXTBSY,
 				   EISDIR, ELIBBAD, EPERM, EINTR,
 				   ECHILD */
+#include <error.h>		/* error() */
 #include <sys/wait.h>		/* waitpid(), WIFEXITED(),
 				   WEXITSTATUS() */
 #include <sysexits.h>		/* EX_OSERR, EX_OSFILE,
@@ -109,7 +110,7 @@
     proc_dir = opendir("/proc");
     if(proc_dir == NULL){
       int e = errno;
-      perror("opendir");
+      error(0, errno, "opendir");
       switch(e){
       case EACCES:
       case ENOTDIR:
@@ -151,7 +152,7 @@
 	char *exe_link;
 	ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  exitstatus = EX_OSERR;
 	  goto failure;
 	}
@@ -165,7 +166,7 @@
 	    continue;
 	  }
 	  int e = errno;
-	  perror("lstat");
+	  error(0, errno, "lstat");
 	  free(exe_link);
 	  switch(e){
 	  case EACCES:
@@ -213,60 +214,60 @@
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
@@ -283,7 +284,7 @@
     goto failure;
   }
   if(splashy_command_pid == -1){
-    perror("fork");
+    error(0, errno, "fork");
     exitstatus = EX_OSERR;
     goto failure;
   }
@@ -293,7 +294,7 @@
       const char splashy_command[] = "/sbin/splashy_update";
       execl(splashy_command, splashy_command, prompt, (char *)NULL);
       int e = errno;
-      perror("execl");
+      error(0, errno, "execl");
       switch(e){
       case EACCES:
       case ENOENT:
@@ -341,7 +342,7 @@
       goto failure;
     }
     if(ret == -1){
-      perror("waitpid");
+      error(0, errno, "waitpid");
       if(errno == ECHILD){
 	splashy_command_pid = 0;
       }
@@ -379,27 +380,27 @@
 	 the real user ID (_mandos) */
       ret = setuid(geteuid());
       if(ret == -1){
-	perror("setuid");
+	error(0, errno, "setuid");
       }
       
       setsid();
       ret = chdir("/");
       if(ret == -1){
-	perror("chdir");
+	error(0, errno, "chdir");
       }
 /*       if(fork() != 0){ */
 /* 	_exit(EXIT_SUCCESS); */
 /*       } */
       ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace stdout */
       if(ret == -1){
-	perror("dup2");
+	error(0, errno, "dup2");
 	_exit(EX_OSERR);
       }
       
       execl("/sbin/splashy", "/sbin/splashy", "boot", (char *)NULL);
       {
 	int e = errno;
-	perror("execl");
+	error(0, errno, "execl");
 	switch(e){
 	case EACCES:
 	case ENOENT:
@@ -425,13 +426,13 @@
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());

=== modified file 'plugins.d/usplash.c'
--- plugins.d/usplash.c	2010-03-27 18:39:02 +0000
+++ plugins.d/usplash.c	2010-09-07 16:48:58 +0000
@@ -31,19 +31,20 @@
 #include <fcntl.h>		/* open(), O_WRONLY, O_RDONLY */
 #include <iso646.h>		/* and, or, not*/
 #include <errno.h>		/* errno, EINTR */
+#include <error.h>
 #include <sys/types.h>		/* size_t, ssize_t, pid_t, DIR, struct
 				   dirent */
 #include <stddef.h>		/* NULL */
 #include <string.h>		/* strlen(), memcmp() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf()*/
 #include <unistd.h>		/* close(), write(), readlink(),
 				   read(), STDOUT_FILENO, sleep(),
 				   fork(), setuid(), geteuid(),
 				   setsid(), chdir(), dup2(),
 				   STDERR_FILENO, execv() */
 #include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
-				   EXIT_SUCCESS, malloc(), _exit() */
-#include <stdlib.h>		/* getenv() */
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
 #include <dirent.h>		/* opendir(), readdir(), closedir() */
 #include <inttypes.h>		/* intmax_t, strtoimax() */
 #include <sys/stat.h>		/* struct stat, lstat(), S_ISLNK */
@@ -153,7 +154,7 @@
   size_t cmdline_len = 0;
   DIR *proc_dir = opendir("/proc");
   if(proc_dir == NULL){
-    perror("opendir");
+    error(0, errno, "opendir");
     return -1;
   }
   errno = 0;
@@ -181,7 +182,7 @@
       char *exe_link;
       ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
       if(ret == -1){
-	perror("asprintf");
+	error(0, errno, "asprintf");
 	goto fail_find_usplash;
       }
       
@@ -193,7 +194,7 @@
 	  free(exe_link);
 	  continue;
 	}
-	perror("lstat");
+	error(0, errno, "lstat");
 	free(exe_link);
 	goto fail_find_usplash;
       }
@@ -224,13 +225,13 @@
 	ret = asprintf(&cmdline_filename, "/proc/%s/cmdline",
 		       proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  goto fail_find_usplash;
 	}
 	cl_fd = open(cmdline_filename, O_RDONLY);
 	free(cmdline_filename);
 	if(cl_fd == -1){
-	  perror("open");
+	  error(0, errno, "open");
 	  goto fail_find_usplash;
 	}
       }
@@ -242,7 +243,7 @@
 	if(cmdline_len + blocksize > cmdline_allocated){
 	  tmp = realloc(cmdline, cmdline_allocated + blocksize);
 	  if(tmp == NULL){
-	    perror("realloc");
+	    error(0, errno, "realloc");
 	    close(cl_fd);
 	    goto fail_find_usplash;
 	  }
@@ -253,7 +254,7 @@
 	sret = read(cl_fd, cmdline + cmdline_len,
 		    cmdline_allocated - cmdline_len);
 	if(sret == -1){
-	  perror("read");
+	  error(0, errno, "read");
 	  close(cl_fd);
 	  goto fail_find_usplash;
 	}
@@ -261,14 +262,14 @@
       } while(sret != 0);
       ret = close(cl_fd);
       if(ret == -1){
-	perror("close");
+	error(0, errno, "close");
 	goto fail_find_usplash;
       }
     }
     /* Close directory */
     ret = closedir(proc_dir);
     if(ret == -1){
-      perror("closedir");
+      error(0, errno, "closedir");
       goto fail_find_usplash;
     }
     /* Success */
@@ -323,26 +324,26 @@
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -351,7 +352,7 @@
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -360,7 +361,7 @@
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -369,7 +370,7 @@
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -378,7 +379,7 @@
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -387,7 +388,7 @@
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -399,7 +400,7 @@
   /* Write command to FIFO */
   if(not usplash_write(&fifo_fd, "TIMEOUT", "0")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -411,7 +412,7 @@
   
   if(not usplash_write(&fifo_fd, "INPUTQUIET", prompt)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -429,7 +430,7 @@
   outfifo_fd = open("/dev/.initramfs/usplash_outfifo", O_RDONLY);
   if(outfifo_fd == -1){
     if(errno != EINTR){
-      perror("open");
+      error(0, errno, "open");
       status = EX_OSERR;
     }
     goto failure;
@@ -448,7 +449,7 @@
       char *tmp = realloc(buf, buf_allocated + blocksize);
       if(tmp == NULL){
 	if(errno != EINTR){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -460,7 +461,7 @@
 		buf_allocated - buf_len);
     if(sret == -1){
       if(errno != EINTR){
-	perror("read");
+	error(0, errno, "read");
 	status = EX_OSERR;
       }
       TEMP_FAILURE_RETRY(close(outfifo_fd));
@@ -475,7 +476,7 @@
   ret = close(outfifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -488,7 +489,7 @@
   
   if(not usplash_write(&fifo_fd, "TIMEOUT", "15")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -501,7 +502,7 @@
   ret = close(fifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -515,7 +516,7 @@
       sret = write(STDOUT_FILENO, buf + written, buf_len - written);
       if(sret == -1){
 	if(errno != EINTR){
-	  perror("write");
+	  error(0, errno, "write");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -552,7 +553,7 @@
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -561,7 +562,7 @@
   if(outfifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(outfifo_fd));
     if(ret == -1){
-      perror("close");
+      error(0, errno, "close");
     }
   }
   
@@ -575,7 +576,7 @@
 			   (sizeof(char *)
 			    * (size_t)(cmdline_argc + 2)));
       if(tmp == NULL){
-	perror("realloc");
+	error(0, errno, "realloc");
 	free(cmdline_argv);
 	return status;
       }
@@ -602,23 +603,27 @@
        the real user ID (_mandos) */
     ret = setuid(geteuid());
     if(ret == -1){
-      perror("setuid");
+      error(0, errno, "setuid");
     }
     
     setsid();
     ret = chdir("/");
+    if(ret == -1){
+      error(0, errno, "chdir");
+      _exit(EX_OSERR);
+    }
 /*     if(fork() != 0){ */
 /*       _exit(EXIT_SUCCESS); */
 /*     } */
     ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
     if(ret == -1){
-      perror("dup2");
+      error(0, errno, "dup2");
       _exit(EX_OSERR);
     }
     
     execv(usplash_name, cmdline_argv);
     if(not interrupted_by_signal){
-      perror("execv");
+      error(0, errno, "execv");
     }
     free(cmdline);
     free(cmdline_argv);
@@ -629,7 +634,7 @@
   sleep(2);
   if(not usplash_write(&fifo_fd, "PULSATE", NULL)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
     }
   }
   
@@ -637,7 +642,7 @@
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -648,13 +653,13 @@
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());

